import abc
import asyncio
import base64
import builtins
import datetime
import functools
import hashlib
import importlib
import inspect
import json
from decimal import Decimal
from enum import Enum
from functools import cached_property
from functools import lru_cache
from functools import partial
from pathlib import Path
from pathlib import PosixPath
from types import NoneType
from typing import Any
from typing import Callable
from typing import Dict
from typing import ForwardRef
from typing import Hashable
from typing import Iterable
from typing import List
from typing import Mapping
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union
from typing import cast
from typing import get_origin
from uuid import UUID

import anyio
import attr
from cachetools import LRUCache
from cattrs import Converter
from cattrs._compat import is_generic
from cattrs.gen import make_dict_unstructure_fn
from cattrs.gen import override
from httpx import URL
from humps import camelize
from pydantic import BaseModel
from pydantic_core import PydanticUndefined

from imbue_core.errors import ImbueError
from imbue_core.fixed_traceback import FixedTraceback
from imbue_core.frozen_utils import FrozenDict
from imbue_core.frozen_utils import FrozenMapping
from imbue_core.serialization import SerializedException
from imbue_core.serialization_types import Serializable

T = TypeVar("T")
TYPE_KEY = "__type"
EXCEPTION_KEY = "__exception"

# LABELS for marking attributes with special handling
DONT_SERIALIZE_METADATA_KEY = "_imbue_dont_serialize"
DONT_SERIALIZE = {DONT_SERIALIZE_METADATA_KEY: True}
SERIALIZE_WITH_DEFAULT_KEY = "_imbue_serialize_with_default"
SERIALIZE_WITH_DEFAULT = {SERIALIZE_WITH_DEFAULT_KEY: True}

SERIALIZABLE_PROPERTY_KEY = "_imbue_is_serializable_property"
CACHED_SERIALIZABLE_PROPERTY_KEY = "_imbue_is_cached_serializable_property"


##########################################################################################
# UTILITY FUNCTIONS
##########################################################################################


def _safe_issubclass(t1: type, t2: type) -> bool:
    return inspect.isclass(t1) and issubclass(t1, t2)


def _is_frozen_mapping_type(t: type) -> bool:
    return _safe_issubclass(get_origin(t) or t, FrozenMapping)


def _is_mapping_type(t: type) -> bool:
    return _safe_issubclass(get_origin(t) or t, Mapping)


_ALLOWED_SPECIAL_MAPPING_TYPES = (LRUCache,)


def _is_special_mapping_type(t: type) -> bool:
    return t in _ALLOWED_SPECIAL_MAPPING_TYPES


def _is_str_type_special_mapping_type(t: str) -> bool:
    return t in [_type_to_string(t, fully_qualified=True) for t in _ALLOWED_SPECIAL_MAPPING_TYPES]


def _is_obj_supported_primitive(obj: Any) -> bool:
    return type(obj) in {bool, int, float, str, NoneType}


def _type_to_string(t: type, fully_qualified: bool) -> str:
    name = t.__name__
    if fully_qualified:
        return f"{t.__module__}.{name}"
    else:
        return name


def _type_from_string(type_str: str) -> Any:
    if "[" in type_str:
        class_details, _ = type_str.split("[", 1)
    else:
        class_details = type_str
    if "." in class_details:
        module_path, class_name = class_details.rsplit(".", 1)
        module = importlib.import_module(module_path)
    else:
        class_name = class_details
        module = builtins
    result = getattr(module, class_name)
    return result


def calculate_object_content_hash(obj: Any, keys_to_ignore: Iterable[str]) -> str:
    # NOTE: if your obj uses @serializable_property or @cached_serializable_property, those will
    #  be included in the hash, so be careful. If you don't want them to affect the hash, either
    #  a) don't use @serializable_property or @cached_serializable_property, or
    #  b) add those properties to the keys_to_ignore.
    md5_hash = hashlib.md5()
    md5_hash.update(
        serialize_to_json(
            {
                k: getattr(obj, k) if k != TYPE_KEY else obj.__class__.__name__
                for k in serialize_to_dict(obj).keys()
                if k not in keys_to_ignore
            }
        ).encode()
    )
    return md5_hash.hexdigest()


def get_serializable_properties(obj: Any) -> Dict[str, Any]:
    members = inspect.getmembers(type(obj))
    marked_members = {}
    for name, member in members:
        if is_serializable_property(member):
            marked_members[name] = getattr(obj, name)
    return marked_members


def is_serializable_property(func: Callable) -> bool:
    return getattr(func, CACHED_SERIALIZABLE_PROPERTY_KEY, False) or (
        isinstance(func, property) and getattr(func.fget, SERIALIZABLE_PROPERTY_KEY, False)
    )


def cached_serializable_property(func: Callable[..., T]) -> cached_property[T]:
    property_to_return = cached_property(func)
    setattr(property_to_return, CACHED_SERIALIZABLE_PROPERTY_KEY, True)
    return property_to_return


def serializable_property(func: Callable[..., T]) -> property:
    property_to_return = func
    # NOTE: this will be stored in the fget attribute of the property, which is also the function
    #   we are decorating, so we must check in `func.fget` to see if the property is serializable.
    #   We need to do it this way because we cannot set the attribute on the property object/wrapper
    #   itself, because of the way the inbuilt `property` decorator works.
    setattr(property_to_return, SERIALIZABLE_PROPERTY_KEY, True)
    return property(property_to_return)


def get_dont_serialize_member_names_of_type(obj_type: type) -> List[str]:
    if not attr.has(obj_type):
        return []
    return [field.name for field in attr.fields(obj_type) if field.metadata.get(DONT_SERIALIZE_METADATA_KEY, False)]


def get_serialize_with_default_member_names_of_type(obj_type: type) -> Mapping[str, Any]:
    if _safe_issubclass(obj_type, BaseModel):
        model_fields = getattr(obj_type, "model_fields", {})
        return {
            name: None if field.default == PydanticUndefined else field.default for name, field in model_fields.items()
        }
    if not attr.has(obj_type):
        return {}
    return {
        field.name: None if field.default == attr.NOTHING else field.default
        for field in attr.fields(obj_type)
        if field.metadata.get(SERIALIZE_WITH_DEFAULT_KEY, False)
    }


def get_dont_serialize_member_names(obj: Any) -> List[str]:
    if not attr.has(obj):
        return []
    members = inspect.getmembers(obj)
    marked_members = []
    for name, _ in members:
        if is_dont_serialize_member(obj, name):
            marked_members.append(name)
    return marked_members


def is_dont_serialize_member(obj: Any, member_name: str) -> bool:
    if not attr.has(obj):
        return False
    for field in attr.fields(obj.__class__):  # type: ignore
        if field.name == member_name:
            return bool(field.metadata.get(DONT_SERIALIZE_METADATA_KEY, False))
    return False


class SerializationError(ImbueError):
    """Raised when we encounter problems related to Serialization or Deserialization."""


def _to_json_dumpable_object_without_type_keys(data: Any) -> Any:
    if isinstance(data, dict):
        if data.get(TYPE_KEY, "") in {
            _type_to_string(PosixPath, fully_qualified=True),
            _type_to_string(Path, fully_qualified=True),
            _type_to_string(UUID, fully_qualified=True),
        }:
            return data["value"]
        else:
            return {
                key: _to_json_dumpable_object_without_type_keys(value)
                for key, value in data.items()
                if key != TYPE_KEY
            }
    elif isinstance(data, list):
        return [_to_json_dumpable_object_without_type_keys(item) for item in data]
    elif _is_obj_supported_primitive(data):
        return data
    else:
        return str(data)


def _camelize_keys_which_represent_python_names(data: Any) -> Any:
    """Converts JSON-style objects to use camel case keys.

    Takes a JSON-style object produced by CONVERTER.structure and returns the same object with certain
    keys converted to camel case. Camel cases keys which are derived from names of Python attributes and properties.
    Does not camel-case keys which were keys of dictionaries before serialization.

    See cattrs_serialization_test.test_camel_casing for an example.
    """
    if isinstance(data, dict):
        if TYPE_KEY not in data or issubclass(_type_from_string(data[TYPE_KEY]), Mapping):
            return {key: _camelize_keys_which_represent_python_names(value) for key, value in data.items()}
        else:
            return {camelize(key): _camelize_keys_which_represent_python_names(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [_camelize_keys_which_represent_python_names(item) for item in data]
    else:
        return data


##########################################################################################
# CLASS-SPECIFIC HOOKS
##########################################################################################


class _ShouldDeserialize:
    pass


# FIXME: Types such as LRUCache will always serialize without errors since they inherit from Mapping but will not deserialize correctly.
#  We should either document this behavior or change it so that the serialization fails if the type is not supported.
def _serialize_mapping_to_json_dict(data: Mapping, converter: Converter) -> Any:
    assert _is_mapping_type(type(data)), f"Attempted to serialize object of type {type(data)} as a mapping."
    return {str(converter.unstructure(k)): converter.unstructure(v) for k, v in data.items()}


def _serialize_mapping(data: Mapping, converter: Converter) -> Any:
    assert _is_mapping_type(type(data)), f"Attempted to serialize object of type {type(data)} as a mapping."
    entries = [(converter.unstructure(k), converter.unstructure(v)) for k, v in data.items()]
    return {TYPE_KEY: _type_to_string(type(data), fully_qualified=True), "__entries": entries}


def _deserialize_special_mapping_types(data: Dict, type_key: str) -> Mapping:
    if type_key == _type_to_string(LRUCache, fully_qualified=True):
        # FIXME: We're not serializing the object correctly and so the deserialization is hacky
        obj: LRUCache = LRUCache(maxsize=10000)
        return obj
    else:
        raise ValueError(f"Unsupported type {type_key}")


def _deserialize_mapping(data: Dict, mapping_type: type, converter: Converter) -> Mapping:
    if TYPE_KEY in data and _is_str_type_special_mapping_type(data[TYPE_KEY]):
        return _deserialize_special_mapping_types(data, data[TYPE_KEY])

    out = {}
    if "__entries" in data:
        entries = data["__entries"]
    else:
        # We keep this branch for backwards compatibility with mappings serialized as dictionaries.
        # We do not support Yasoo's DictWithSerializedKeys -- those will need to be migrated to the new format.
        if TYPE_KEY in data:
            del data[TYPE_KEY]
        entries = data.items()

    for k, v in entries:
        out[converter.structure(k, _ShouldDeserialize)] = converter.structure(v, _ShouldDeserialize)

    if _is_frozen_mapping_type(mapping_type):
        return FrozenDict(out)
    return out


def _serialize_frozen_set(data: frozenset, converter: Converter) -> Dict:
    assert type(data) is frozenset, f"Attempted to serialize object of type {type(data)} as a frozenset."
    value = converter.unstructure(data, unstructure_as=list)
    return {"value": value, TYPE_KEY: _type_to_string(type(data), fully_qualified=True)}


def _deserialize_frozen_set(data: Dict, _: type, converter: Converter) -> frozenset:
    return frozenset(converter.structure(data["value"], list))


def _serialize_uuid(data: UUID) -> Dict:
    if type(data) is UUID:
        return {"value": data.hex, TYPE_KEY: _type_to_string(type(data), fully_qualified=True)}
    elif type(data) is str:
        return {"value": data, TYPE_KEY: _type_to_string(UUID, fully_qualified=True)}
    else:
        raise TypeError("Tried to serialize " + str(data) + ", which is neither a string nor a UUID, as a UUID.")


def _deserialize_uuid(data: Dict[str, str] | str, _: type) -> UUID:
    if isinstance(data, dict):
        return UUID(data["value"])
    elif isinstance(data, str):
        return UUID(data)
    else:
        raise TypeError("Tried to deserialize something which is neither a string nor a dictionary, as a UUID.")


def _serialize_tuple(data: tuple, converter: Converter) -> Dict:
    assert type(data) is tuple, f"Attempted to serialize object of type {type(data)} as a tuple."
    return {
        "value": [converter.unstructure(x) for x in data],
        TYPE_KEY: _type_to_string(type(data), fully_qualified=True),
    }


def _deserialize_tuple(data: Dict, _: type, converter: Converter) -> tuple:
    return tuple(converter.structure(x, _ShouldDeserialize) for x in data["value"])


def _serialize_url(data: URL) -> Dict:
    assert type(data) is URL, f"Tried to serialize {data} which is not a URL."
    return {"value": str(data), TYPE_KEY: _type_to_string(type(data), fully_qualified=True)}


def _deserialize_url(data: Dict, _: type) -> URL:
    return URL(data["value"])


def _serialize_decimal(data: Decimal) -> Dict:
    assert type(data) is Decimal, f"Attempted to serialize object of type {type(data)} as a Decimal."
    return {"value": str(data), TYPE_KEY: _type_to_string(type(data), fully_qualified=True)}


def _deserialize_decimal(data: Dict, _: type) -> Decimal:
    return Decimal(data["value"])


def _serialize_traceback(data: FixedTraceback) -> Dict:
    assert _safe_issubclass(
        type(data), FixedTraceback
    ), f"Attempted to serialize object of type {type(data)} as a traceback."
    return {"value": data.to_dict(), TYPE_KEY: _type_to_string(type(data), fully_qualified=True)}


def _deserialize_traceback(data: Dict, _: type) -> FixedTraceback:
    return FixedTraceback.from_dict(data["value"])


def _serialize_path(data: Path) -> Dict:
    assert _safe_issubclass(type(data), Path), f"Attempted to serialize an object of type {type(data)} as a Path."
    return {"value": str(data), TYPE_KEY: _type_to_string(type(data), fully_qualified=True)}


def _deserialize_path(data: Any, _: type) -> Path:
    if type(data) is dict:
        return Path(data["value"])
    return Path(data)


def _serialize_anyio_path(data: anyio.Path) -> Dict:
    assert _safe_issubclass(
        type(data), anyio.Path
    ), f"Attempted to serialize an object of type {type(data)} as a Path."
    return {"value": str(data), TYPE_KEY: _type_to_string(type(data), fully_qualified=True)}


def _deserialize_anyio_path(data: Any, _: type) -> anyio.Path:
    if type(data) is dict:
        return anyio.Path(data["value"])
    return anyio.Path(data)


def _serialize_datetime(data: datetime.datetime) -> Dict:
    assert _safe_issubclass(
        type(data), datetime.datetime
    ), f"Attempted to serialize object of type {type(data)} as a datetime."
    return {
        TYPE_KEY: _type_to_string(type(data), fully_qualified=True),
        "time": data.astimezone(datetime.timezone.utc).timestamp(),
        "tzaware": data.tzinfo is not None,
    }


def _deserialize_datetime(data: Dict, _: type) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(data["time"], datetime.timezone.utc if data.get("tzaware", None) else None)


def _serialize_bytes(data: bytes) -> Dict:
    assert type(data) is bytes, f"Attempted to serialize object of type {type(data)} as bytes."
    return {
        TYPE_KEY: _type_to_string(type(data), fully_qualified=True),
        # use ascii since base64 guarantees ascii characters only
        "value": base64.b64encode(data).decode("ascii"),
    }


def _deserialize_bytes(data: Dict, _: type) -> bytes:
    return base64.b64decode(data["value"])


def _is_forward_ref(t: type) -> bool:
    return isinstance(t, ForwardRef)


def _serialize_forward_ref(data: Any, converter: Converter) -> Any:
    return converter.unstructure(data, unstructure_as=type(data))


def _deserialize_forward_ref(data: Any, _: type, converter: Converter) -> Any:
    # TODO: think of a way to evaluate the ForwardRef _, to improve type safety.
    #  Once we do that, we can swap out the evaluated type for ShouldDeserialize
    #  and enforce that we're getting an object of the correct type.
    return _deserialize_serialized_object(data, _ShouldDeserialize, converter)


def _is_union_type(t: type) -> bool:
    return get_origin(t) is Union


def _deserialize_union_type(data: Any, type_of_data: type, converter: Converter) -> Any:
    return converter.structure(data, _ShouldDeserialize)


def _serialize_enum(data: Enum, converter: Converter) -> Any:
    assert inspect.isclass(type(data)) and issubclass(
        type(data), Enum
    ), f"Attempted to serialize object of type {type(data)} as an Enum."
    return converter._unstructure_enum(data)


def _deserialize_enum(data: Dict[str, str] | str, t: type) -> Any:
    # We include this complicated logic to preserve backwards compatibility with old JSON that was
    # serialized by Yasoo. Yasoo serialized enums by converting them into the form
    # {"__type": "...", "value": "..."}. Strangely, Yasoo converted this dictionary into a string
    # whenever an enum value occurred as a dictionary key, but did not convert it into a string
    # when it occurred anywhere else. Hence we need to handle enums that are represented by
    # dictionaries, stringified dictionaries, and strings.

    assert _safe_issubclass(t, Enum)

    if isinstance(data, str):
        try:
            # This is the case where data is an enum value, serialized by Cattrs.
            return t(data)
        except ValueError:
            # This is the case where data is a stringified dictionary, serialized by Yasoo.
            data_as_dict = json.loads(data)
            return t[data_as_dict["value"]]  # type: ignore
    else:
        # This is the case where data is a dictionary, serialized by Yasoo.
        return t[data["value"]]  # type: ignore


##########################################################################################
# TYPE KEY LOGIC
##########################################################################################


class _AvoidTypeKeyLogic:
    pass


@lru_cache
def flag_to_ignore_type_key_hooks(t: type) -> type:
    class GivenTypeFlaggedToAvoidTypeKeyLogic(t, _AvoidTypeKeyLogic):
        pass

    GivenTypeFlaggedToAvoidTypeKeyLogic.__name__ = t.__name__
    GivenTypeFlaggedToAvoidTypeKeyLogic.__qualname__ = t.__qualname__

    return GivenTypeFlaggedToAvoidTypeKeyLogic


def get_pydantic_model_attributes(model: BaseModel) -> Dict[str, Any]:
    # This is a hack to dump only the top level but also avoid dumping any properties
    attributes = getattr(type(model), "model_fields", {})
    return {a: getattr(model, a) for a in attributes}


# These two factory functions produce the functions for serializing attr classes.
# Only one of them should be registered at a time, depending on whether we are including
# do-not-serialize fields in the serialization.
def _serialize_attr_class_factory(cls: type, converter: Converter) -> Callable[[Any], Any]:
    return make_dict_unstructure_fn(cls, converter)


def _serialize_attr_class_without_dont_serialize_fields(
    cls: type, converter: Converter, is_camel_case: bool
) -> Callable[[Any], Any]:
    members_to_omit = get_dont_serialize_member_names_of_type(cls)
    omit_kwargs = {name: override(omit=True) for name in members_to_omit}
    return make_dict_unstructure_fn(cls, converter, **omit_kwargs)  # type: ignore


def _serialize_with_type_key(data: Any, converter: Converter, for_javascript: bool = False) -> Any:
    type_of_data = type(data)

    if _is_obj_supported_primitive(data) or isinstance(data, list) or isinstance(data, tuple):
        # This means that data was annotated as a Serializable, but it is a primitive or a tuple.
        return converter.unstructure(data, unstructure_as=type_of_data)

    type_of_data_with_typekey_already_added = flag_to_ignore_type_key_hooks(type_of_data)  # type: ignore

    # This is a hack which is necessary because cattrs does not work well with Protocols.
    # Protocols are generic classes, but they don't have __orig_bases__, which cattrs
    # assumes them to have.
    if is_generic(type_of_data_with_typekey_already_added):
        old_orig_bases = getattr(type_of_data_with_typekey_already_added, "__orig_bases__", ())
        setattr(type_of_data_with_typekey_already_added, "__orig_bases__", old_orig_bases)

    if isinstance(data, BaseModel):
        # This is a shortcut: when you encounter a Pydantic model, just use Pydantic serialization.
        # NOTE: currently we don't support `DONT_SERIALIZE` fields in pydantic models.
        #  so we just serialize all fields.
        unstructured = data.model_dump(by_alias=for_javascript, mode="json")
    else:
        unstructured = converter.unstructure(data, unstructure_as=type_of_data_with_typekey_already_added)

    assert isinstance(unstructured, dict)

    if for_javascript:
        unstructured.update({k: converter.unstructure(v) for k, v in get_serializable_properties(data).items()})

    return {
        TYPE_KEY: _type_to_string(type_of_data, fully_qualified=True),
        **unstructured,
    }


# This is the predicate used in the factory functions above, so they trigger for serializable and attr classes
# that have had their type key logic handled.
def _should_serialize_without_type_key(t: type) -> bool:
    is_serializable_class = _safe_issubclass(t, Serializable) or attr.has(t) or _safe_issubclass(t, BaseModel)
    return is_serializable_class and _safe_issubclass(t, _AvoidTypeKeyLogic)


def _should_add_type_key(t: type) -> bool:
    is_serializable_class = _safe_issubclass(t, Serializable) or attr.has(t) or _safe_issubclass(t, BaseModel)
    return is_serializable_class and not _safe_issubclass(t, _AvoidTypeKeyLogic)


def _deserialize_serialized_object(data: Any, type_of_data: type, converter: Converter) -> Any:
    if isinstance(data, List):
        # Data is a list of objects.
        return converter.structure(data, List[_ShouldDeserialize])
    elif not isinstance(data, Mapping):
        # Data is a primitive, like an integer or a string.
        return converter.structure(data, type(data))
    else:
        # Data is a dictionary with a type key, representing an attrs object, a Pydantic model, or a Mapping
        return _deserialize_using_type_marker(data, type_of_data, converter)


def _should_deserialize_with_type_key_logic(t: type) -> bool:
    is_type_that_should_be_deserialized = (
        attr.has(t)
        or _safe_issubclass(t, Serializable)
        or _safe_issubclass(t, _ShouldDeserialize)
        or t is Hashable
        or _is_mapping_type(t)
        or _safe_issubclass(t, BaseModel)
    )
    should_avoid_type_key_logic = _safe_issubclass(t, _AvoidTypeKeyLogic) or _safe_issubclass(
        get_origin(t) or NoneType, _AvoidTypeKeyLogic
    )
    return is_type_that_should_be_deserialized and not should_avoid_type_key_logic


def deserialized_object_violates_target_type(obj: Any, target_type: type) -> bool:
    if target_type is _ShouldDeserialize or target_type is Serializable:
        return False
    if type(target_type) is TypeVar:
        # We're not really able to check if the object is an instance of a type that's behind a TypeVar.
        return False
    return not isinstance(obj, get_origin(target_type) or target_type)


# Note that expected_type_based_on_annotations may be much more vague than the actual type of the object.
# For example: it may be Serializable, when the object is supposed to be
# deserialized as a HammerResult. We get the real type from the "__type" key.
def _deserialize_using_type_marker(
    obj: Mapping[Any, Any], expected_type_based_on_annotations: Type[T], converter: Converter
) -> T:
    if TYPE_KEY in obj:
        type_of_obj = _type_from_string(obj[TYPE_KEY])
    else:
        type_of_obj = expected_type_based_on_annotations

    if _is_special_mapping_type(type_of_obj):
        pass
    elif _is_frozen_mapping_type(type_of_obj):
        obj.pop(TYPE_KEY, None)  # type: ignore
        type_of_obj = FrozenMapping[_ShouldDeserialize, _ShouldDeserialize]
    elif _is_mapping_type(type_of_obj):
        obj.pop(TYPE_KEY, None)  # type: ignore
        type_of_obj = dict[_ShouldDeserialize, _ShouldDeserialize]
    elif _safe_issubclass(type_of_obj, BaseModel):
        assert isinstance(obj, dict)
        obj.pop(TYPE_KEY, None)
        return cast(T, type_of_obj.model_validate(obj))
    elif not attr.has(type_of_obj):
        # This happens when there is a primitive object which is annotated as Serializable.
        return converter.structure(obj, type_of_obj)  # type: ignore

    # By mixing in the "avoid type key logic" class, force cattrs to do its normal behavior.
    ret: T = converter.structure(obj, flag_to_ignore_type_key_hooks(type_of_obj))

    if inspect.isclass(type_of_obj):
        # Upcast the result so that it has the correct type again, without the mixin.
        object.__setattr__(ret, "__class__", type_of_obj)

    if deserialized_object_violates_target_type(ret, expected_type_based_on_annotations):
        raise TypeError(
            f"Tried to deserialize into type {expected_type_based_on_annotations}, but got object of type {type(ret)}"
        )

    return ret


def _resolve_default(default: Any) -> Any:
    if isinstance(default, attr.Factory):  # type: ignore
        return default.factory()
    return default


def _serialize_with_defaults(cls: type, converter: Converter) -> Callable[[Any], Any]:
    # Handle a pydantic model
    if _safe_issubclass(cls, BaseModel):
        return lambda x: {k: converter.unstructure(v) for k, v in get_pydantic_model_attributes(x).items()}

    members_with_defaults = get_serialize_with_default_member_names_of_type(cls)
    overriden_kwargs = {
        name: override(unstruct_hook=(lambda _, value=_resolve_default(default): value))  # type: ignore
        for name, default in members_with_defaults.items()
    }
    return make_dict_unstructure_fn(cls, converter, **overriden_kwargs)  # type: ignore


def _should_serialize_as_serialized_exception(t: type) -> bool:
    return (
        _safe_issubclass(get_origin(t) or t, BaseException) and not attr.has(t) and not _safe_issubclass(t, BaseModel)
    )


##########################################################################################
# CONVERTER FACTORY
##########################################################################################


class _ConverterFactory:
    """Factory for creating converters with different configurations.

    e.g. for serializing to javascript, or python, or to include do-not-serialize fields.
    """

    def build_base_converter(self) -> Converter:
        # Builds of new base converter object, which registers all the hooks that are common to all converters.
        # The idea being that all new converters start from this base and then override hooks they need to change
        # NOTE: we need to generate a new converter object for each independent concrete converter (as opposed to
        # using converter.copy()) since we use partial functions/closures and this way we ensure the function is
        # being called with the correct converter object.
        converter = Converter()

        converter.register_structure_hook_func(_is_mapping_type, partial(_deserialize_mapping, converter=converter))
        # serialization of mapping types depends on the specific converter so is done in the get_converter factory method

        converter.register_unstructure_hook(frozenset, partial(_serialize_frozen_set, converter=converter))
        converter.register_structure_hook(frozenset, partial(_deserialize_frozen_set, converter=converter))

        converter.register_unstructure_hook(UUID, _serialize_uuid)
        converter.register_structure_hook(UUID, _deserialize_uuid)

        converter.register_unstructure_hook(URL, _serialize_url)
        converter.register_structure_hook(URL, _deserialize_url)

        converter.register_unstructure_hook(Decimal, _serialize_decimal)
        converter.register_structure_hook(Decimal, _deserialize_decimal)

        converter.register_unstructure_hook(FixedTraceback, _serialize_traceback)
        converter.register_structure_hook(FixedTraceback, _deserialize_traceback)

        converter.register_unstructure_hook(Path, _serialize_path)
        converter.register_structure_hook(Path, _deserialize_path)

        converter.register_unstructure_hook(anyio.Path, _serialize_anyio_path)
        converter.register_structure_hook(anyio.Path, _deserialize_anyio_path)

        converter.register_unstructure_hook(datetime.datetime, _serialize_datetime)
        converter.register_structure_hook(datetime.datetime, _deserialize_datetime)

        converter.register_unstructure_hook(bytes, _serialize_bytes)
        converter.register_structure_hook(bytes, _deserialize_bytes)

        converter.register_unstructure_hook(PosixPath, _serialize_path)
        converter.register_structure_hook(PosixPath, _deserialize_path)

        converter.register_unstructure_hook_func(_is_forward_ref, partial(_serialize_forward_ref, converter=converter))
        converter.register_structure_hook_func(_is_forward_ref, partial(_deserialize_forward_ref, converter=converter))

        converter.register_structure_hook_func(_is_union_type, partial(_deserialize_union_type, converter=converter))

        converter.register_structure_hook(NoneType, lambda data, _: None)

        converter.register_unstructure_hook(Enum, partial(_serialize_enum, converter=converter))
        converter.register_structure_hook(Enum, _deserialize_enum)

        converter.register_unstructure_hook_func(
            _should_serialize_as_serialized_exception,
            lambda e: serialize_to_dict(SerializedException.build(e), use_defaults_for_unserializable_fields=True),
        )

        converter.register_structure_hook_func(
            _should_deserialize_with_type_key_logic,
            partial(_deserialize_serialized_object, converter=converter),
        )

        converter.register_structure_hook_func(
            lambda t: isinstance(t, TypeVar), partial(_deserialize_serialized_object, converter=converter)
        )

        return converter

    def get_converter_with_defaults(self, converter: Converter) -> Converter:
        converter.register_unstructure_hook(asyncio.Lock, lambda _: None)
        converter.register_structure_hook(asyncio.Lock, lambda data, _: asyncio.Lock())

        converter.register_unstructure_hook(asyncio.Task, lambda _: None)
        converter.register_structure_hook(asyncio.Task, lambda data, _: None)

        converter.register_unstructure_hook(asyncio.Queue, lambda _: None)
        converter.register_structure_hook(asyncio.Queue, lambda data, _: None)

        converter.register_unstructure_hook(asyncio.Event, lambda _: None)
        converter.register_structure_hook(asyncio.Event, lambda data, _: None)

        converter.register_unstructure_hook(asyncio.Semaphore, lambda _: None)
        converter.register_structure_hook(asyncio.Semaphore, lambda data, _: None)

        converter.register_unstructure_hook(abc.ABCMeta, lambda _: None)
        converter.register_structure_hook(abc.ABCMeta, lambda data, _: None)
        converter.register_unstructure_hook_factory(
            _should_serialize_without_type_key, partial(_serialize_with_defaults, converter=converter)
        )

        return converter

    @functools.cache
    def get_converter(
        self,
        for_javascript: bool = False,
        exclude_dont_serialize_fields: bool = False,
        use_defaults_for_unserializable_fields: bool = False,
    ) -> Converter:
        """Returns a converter with the given configuration.

        The result of this method is cached, so subsequent calls with the same arguments will return the same converter.
        """
        assert not (
            exclude_dont_serialize_fields and use_defaults_for_unserializable_fields
        ), f"Expected exactly one flag to be set, got {exclude_dont_serialize_fields=}, {use_defaults_for_unserializable_fields=}"

        converter = self.build_base_converter()
        if for_javascript:
            converter.register_unstructure_hook_func(
                _is_mapping_type, partial(_serialize_mapping_to_json_dict, converter=converter)
            )
        else:
            converter.register_unstructure_hook_func(
                _is_mapping_type, partial(_serialize_mapping, converter=converter)
            )
            converter.register_unstructure_hook(tuple, partial(_serialize_tuple, converter=converter))
            converter.register_structure_hook(tuple, partial(_deserialize_tuple, converter=converter))

        if exclude_dont_serialize_fields:
            converter.register_unstructure_hook_factory(
                _should_serialize_without_type_key,
                partial(
                    _serialize_attr_class_without_dont_serialize_fields,
                    converter=converter,
                    is_camel_case=for_javascript,
                ),
            )
        else:
            converter.register_unstructure_hook_factory(
                _should_serialize_without_type_key,
                partial(_serialize_attr_class_factory, converter=converter),
            )
        if use_defaults_for_unserializable_fields:
            converter = self.get_converter_with_defaults(converter)

        converter.register_unstructure_hook_func(
            _should_add_type_key, partial(_serialize_with_type_key, converter=converter, for_javascript=for_javascript)
        )

        return converter


CONVERTER_FACTORY = _ConverterFactory()


##########################################################################################
# ENTRY POINTS
##########################################################################################


def _serialize_to_json_dumpable_object(
    obj: Any,
    is_reversible: bool = True,
    for_javascript: bool = False,
    exclude_dont_serialize_fields: bool = False,
    use_defaults_for_unserializable_fields: bool = False,
) -> Any:
    if exclude_dont_serialize_fields:
        # Check and raise error to make it clear to the caller that the object cannot be deserialized.
        # This is a sanity check, to make it easier to debug when using do-not-serialize fields.
        # NOTE: this will only catch cases where non-serializable fields are in obj, but not cases where
        # the non-serializable fields are in nested objects, checking for the nested case is a little complicated
        # so we don't do it basically.
        assert (
            not is_reversible
        ), "Cannot deserialize object when excluding do-not-serialize fields (i.e. when `exclude_dont_serialize_fields=True`). If you want to serialize an object and exclude do-not-serialize fields, make sure to set `is_reversible=False`."

    if use_defaults_for_unserializable_fields:
        # The point of the use_defaults_for_unserializable_fields flag is to make it possible to serialize objects
        # and then recreate them later even if certain fields are not fully saved. We never want to use this flag
        # with `is_reversible=False` since we won't know the type to be able to recreate the object.
        assert is_reversible, "Cannot restructure inputs if is_reversible=False"

    # TODO: this is a hack to make it possible to serialize ExecutionContexts for class method hammers.
    #  This lets us serialize ExecutionContexts for calls to class methods without serializing the class itself.
    #  The long-term solutions are 1) either get rid of all class method hammers,
    #  or 2) write a custom hook that can serialize type objects.
    if type(obj) is dict and "__class__" in obj:
        del obj["__class__"]

    converter = CONVERTER_FACTORY.get_converter(
        for_javascript=for_javascript,
        exclude_dont_serialize_fields=exclude_dont_serialize_fields,
        use_defaults_for_unserializable_fields=use_defaults_for_unserializable_fields,
    )

    dict_result = converter.unstructure(obj)
    if for_javascript:
        dict_result = _camelize_keys_which_represent_python_names(dict_result)

    if not is_reversible:
        return _to_json_dumpable_object_without_type_keys(dict_result)

    return dict_result


def serialize_to_dict(
    obj: Any,
    is_reversible: bool = True,
    for_javascript: bool = False,
    exclude_dont_serialize_fields: bool = False,
    use_defaults_for_unserializable_fields: bool = False,
) -> Dict[str, Any]:
    """Serialize to a python dict."""
    return cast(
        Dict[str, Any],
        _serialize_to_json_dumpable_object(
            obj,
            is_reversible=is_reversible,
            for_javascript=for_javascript,
            exclude_dont_serialize_fields=exclude_dont_serialize_fields,
            use_defaults_for_unserializable_fields=use_defaults_for_unserializable_fields,
        ),
    )


def serialize_to_json(
    obj: Any,
    indent: Optional[int] = None,
    sort_keys: bool = False,
    is_reversible: bool = True,
    for_javascript: bool = False,
    exclude_dont_serialize_fields: bool = False,
    use_defaults_for_unserializable_fields: bool = False,
) -> str:
    """Serialize an object to a JSON string.

    This is the main serialization entrypoint.

    `is_reversible` controls whether we enforce that the result can be deserialized. In some cases we don't care about
    reversibility, e.g. when serializing data for a frontend we often don't care whether we can deserialize.

    `for_javascript` controls whether we use camelCase for keys that originally were Python identifiers.

    `exclude_dont_serialize_fields` controls whether we include do-not-serialize fields in the serialization.
    If this is `False` then any attr class fields marked with as don't serialize, e.g. with `attr.ib(metadata=DONT_SERIALIZE)`,
    will still be included in the serialization. If this is `True` then they will be excluded, however this also means that
    the result will not be reversible (and thus the caller will have to set `is_reversible=False`).

    `use_defaults_for_unserializable_fields` controls whether we fill fields that cannot be serialized with their default values.
    IMPORTANT: If you use this flag, data may be discarded during deserialization.
    The goal is to be able to deserialize fields to the original type without caring about the data contained.
    Default value choices (guided by crafty serialization requirements):
    - Fields that are marked with attr.ib(metadata=SERIALIZE_WITH_DEFAULT) have the following default values:
        - Fields that are marked with `attr.ib(default=...)` or `attr.ib(factory=...)` use their default values.
        - Fields that do not have a default value are filled with None.
    - Asyncio objects are filled with None.
    - Exceptions are replaced with a string representation
    """
    try:
        unstructured = _serialize_to_json_dumpable_object(
            obj,
            is_reversible=is_reversible,
            for_javascript=for_javascript,
            exclude_dont_serialize_fields=exclude_dont_serialize_fields,
            use_defaults_for_unserializable_fields=use_defaults_for_unserializable_fields,
        )
        return json.dumps(unstructured, indent=indent, sort_keys=sort_keys)
    except Exception as e:
        raise SerializationError(str(e)) from e


def deserialize_from_json(
    data: str,
    for_javascript: bool = False,
    exclude_dont_serialize_fields: bool = False,
    use_defaults_for_unserializable_fields: bool = False,
) -> Any:
    try:
        converter = CONVERTER_FACTORY.get_converter(
            for_javascript=for_javascript,
            exclude_dont_serialize_fields=exclude_dont_serialize_fields,
            use_defaults_for_unserializable_fields=use_defaults_for_unserializable_fields,
        )
        return _deserialize_serialized_object(json.loads(data), _ShouldDeserialize, converter=converter)
    except Exception as e:
        raise SerializationError(str(e)) from e


def deserialize_from_dict(
    data: Dict[str, Any],
    as_type: type = _ShouldDeserialize,
    for_javascript: bool = False,
    exclude_dont_serialize_fields: bool = False,
    use_defaults_for_unserializable_fields: bool = False,
) -> Any:
    try:
        converter = CONVERTER_FACTORY.get_converter(
            for_javascript=for_javascript,
            exclude_dont_serialize_fields=exclude_dont_serialize_fields,
            use_defaults_for_unserializable_fields=use_defaults_for_unserializable_fields,
        )
        return _deserialize_using_type_marker(data, as_type, converter=converter)
    except Exception as e:
        raise SerializationError(str(e)) from e


def deserialize_from_dict_with_type(data: Dict[str, Any], obj_type: Type[T]) -> T:
    try:
        converter = CONVERTER_FACTORY.get_converter(for_javascript=False, exclude_dont_serialize_fields=False)
        result = converter.structure(data, obj_type)
        assert isinstance(result, obj_type), f"Expected an object of type {obj_type}, but got {result}"
        return result
    except Exception as e:
        raise SerializationError(str(e)) from e


def deserialize_from_json_with_type(data: Union[str, bytes, bytearray], obj_type: Type[T]) -> T:
    try:
        converter = CONVERTER_FACTORY.get_converter(for_javascript=False, exclude_dont_serialize_fields=False)
        return cast(T, _deserialize_serialized_object(json.loads(data), obj_type, converter=converter))
    except Exception as e:
        raise SerializationError(str(e)) from e



from typing import Callable

from loguru import logger
from tenacity import RetryCallState


def _log_before_sleep(retry_state: RetryCallState, log_fn: Callable[[str], None]) -> None:
    fn_name = retry_state.fn.__name__ if retry_state.fn is not None else "unknown"
    sleep_time = retry_state.next_action.sleep if retry_state.next_action is not None else 0
    outcome = retry_state.outcome
    if outcome is not None:
        exception = outcome.exception()
        error_message = type(exception).__name__ + ": " + str(exception)
    else:
        error_message = "unknown"
    log_fn(
        f"Retrying {fn_name} in {sleep_time:.2f} seconds, attempt {retry_state.attempt_number} after error: {error_message}"
    )


def log_before_sleep(retry_state: RetryCallState) -> None:
    _log_before_sleep(retry_state, logger.debug)


def log_trace_before_sleep(retry_state: RetryCallState) -> None:
    _log_before_sleep(retry_state, logger.trace)



"""This module contains the base errors for the Imbue applications.

Please subclass from one of the errors in this module for errors that you expect other Imbumans to handle.
"""
from typing import Any


class ImbueRuntimeException(BaseException):
    """Base class for all things that could go wrong within Imbue code.

    An ImbueRuntimeException may or may not be recoverable. Most library code should not be throwing
    ImbueRuntimeExceptions.
    """

    _was_logged_by_log_exception: bool

    def __init__(self, *args: Any) -> None:
        super().__init__(*args)
        # New instances start out marked as not-yet-logged.
        self._was_logged_by_log_exception = False


class ImbueError(ImbueRuntimeException, Exception):
    """Base class for all errors that could possibly be handled.

    When you are building the external API of your subcomponent, the errors you throw should subclass from this.
    """


class ExpectedError(ImbueError):
    """Base class for all Imbue errors that we expect to be handled.

    Use this subclass of ImbueError to represent an exception that -must- be handled by a caller. The usual use-case is
    when caller and callee are part of the same subsystem.
    """



from enum import StrEnum
from functools import cached_property
from typing import Annotated
from typing import Dict
from typing import Tuple

from pydantic import Field
from pydantic import Tag
from pydantic import computed_field

from imbue_core.frozen_utils import FrozenDict
from imbue_core.pydantic_serialization import PydanticFrozenDictAnnotation
from imbue_core.pydantic_serialization import SerializableModel
from imbue_core.pydantic_serialization import build_discriminator

ResourceURL = str


class ConflictType(StrEnum):
    MERGE = "MERGE"
    REBASE = "REBASE"
    CHERRY_PICK = "CHERRY_PICK"
    APPLY = "APPLY"
    REVERT = "REVERT"
    BISECT = "BISECT"


class RepoOperation(SerializableModel):
    pass

    @computed_field
    @cached_property
    def is_empty(self) -> bool:
        """Whether this repo operation leaves the repo unchanged.

        Defaults to False. But should be overridden by subclasses as appropriate.
        """
        return False


class ConflictedRepoOperation(RepoOperation):
    object_type: str = "ConflictedRepoOperation"

    blob_content_by_hash: Annotated[FrozenDict[str, bytes], PydanticFrozenDictAnnotation]
    index_content: bytes
    modified_file_contents_by_path: Annotated[FrozenDict[str, bytes], PydanticFrozenDictAnnotation]
    conflict_type: ConflictType
    special_git_file_contents_by_path: Annotated[FrozenDict[str, bytes], PydanticFrozenDictAnnotation]


class CleanRepoOperation(RepoOperation):
    """
    A clean repo operation is a repo operation that has no conflicts.

    It is a contains the staged diff, the unstaged diff, and the combination of the previous two.
    """

    object_type: str = "CleanRepoOperation"
    combined_diff: str
    staged_diff: str = ""
    unstaged_diff: str = ""

    # FIXME: this is now doing validation -- should be converted to the pydantic way of doing this!
    def model_post_init(self, __context) -> None:
        super().model_post_init(__context)
        if self.combined_diff.strip() != "":
            assert (
                self.staged_diff.strip() != "" or self.unstaged_diff.strip() != ""
            ), "combined diff is not empty, so staged and unstaged diffs must be non-empty"

    @computed_field
    @cached_property
    def is_empty(self) -> bool:
        return self.combined_diff.strip() == ""


class RepoState(SerializableModel):
    git_hash: str
    repo_operation: (
        Annotated[CleanRepoOperation, Tag("CleanRepoOperation")]
        | Annotated[ConflictedRepoOperation, Tag("ConflictedRepoOperation")]
    ) = Field(discriminator=build_discriminator())

    @computed_field
    @cached_property
    def is_conflicted(self) -> bool:
        return isinstance(self.repo_operation, ConflictedRepoOperation)

    @computed_field
    @cached_property
    def has_operations(self) -> bool:
        return (
            isinstance(self.repo_operation, ConflictedRepoOperation) or self.repo_operation.combined_diff.strip() != ""
        )

    @computed_field
    @cached_property
    def type_name(self) -> str:
        return self.__class__.__name__

    def build_with_new_commit(self, git_hash: str) -> "RepoState":
        return RepoState(git_hash=git_hash, repo_operation=self.repo_operation)


GIT_FILE_PATH_NAMES_BY_CONFLICT_TYPE: Dict[ConflictType, Tuple[str, ...]] = {
    ConflictType.MERGE: ("MERGE_HEAD", "AUTO_MERGE", "MERGE_MSG", "MERGE_MODE"),
    ConflictType.REBASE: ("REBASE_HEAD",),
    ConflictType.CHERRY_PICK: ("CHERRY_PICK_HEAD",),
    ConflictType.APPLY: (),
    ConflictType.REVERT: ("REVERT_HEAD",),
}



"""Utilities to work with GitLab repositories."""

import os
from contextlib import contextmanager
from typing import Final
from typing import Generator
from typing import Optional

import gitlab
from gitlab.exceptions import GitlabGetError
from loguru import logger
from pydantic import ConfigDict
from pydantic_settings import BaseSettings

from imbue_core.serialization_types import Serializable

GITLAB_TOKEN_NAME: Final[str] = "GITLAB_TOKEN_FOR_USER_REPOS"


class GitlabTokenMissingError(Exception):
    """Raised when the expected environment variable is not set."""


class GitlabProjectReference(BaseSettings, Serializable):
    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True, extra="allow")
    subgroup: str
    project_name: str

    @property
    def path_with_namespace(self) -> str:
        return f"{self.subgroup}/{self.project_name}"


@contextmanager
def gitlab_client_from_environment() -> Generator[gitlab.Gitlab, None, None]:
    """Create a GitLab client from environment variables and manage it as a context.

    Yields:
        gitlab.Gitlab: The configured GitLab client

    Raises:
        GitlabTokenMissingError: If gitlab_token_name environment variable is not set
        GitlabAuthenticationError: If authentication fails
    """
    # Get GitLab token from environment -- this is not a setting, it's a secret.
    gitlab_token = os.getenv(GITLAB_TOKEN_NAME)
    if not gitlab_token:
        error_msg = f"{GITLAB_TOKEN_NAME} environment variable is not set"
        raise GitlabTokenMissingError(error_msg)

    # Get GitLab URL from environment or use default
    gitlab_url = os.getenv("GITLAB_URL", "https://gitlab.com")
    logger.debug(f"Using GitLab URL: {gitlab_url}")

    # Initialize GitLab client
    with gitlab.Gitlab(gitlab_url, private_token=gitlab_token) as gl:
        logger.debug("GitLab client initialized successfully")
        yield gl


def get_gitlab_remote_url(gitlab_token: str, gitlab_project_reference: GitlabProjectReference) -> str:
    """WARNING: This embeds the token in the URL!  Only use this on the server-side backend."""
    # TODO:(CFTY-1244): Perhaps there is a better way to manage this secret?
    return f"https://oauth2:{gitlab_token}@gitlab.com/{gitlab_project_reference.path_with_namespace}.git"


def ensure_gitlab_project_exists(
    gl: gitlab.Gitlab,
    project_reference: GitlabProjectReference,
    visibility: str = "private",
    description: Optional[str] = None,
) -> None:
    path_with_namespace = project_reference.path_with_namespace
    logger.debug(f"Creating GitLab repository: {path_with_namespace=}")

    try:
        found_project = gl.projects.get(path_with_namespace)
        logger.debug(
            f"Repository already exists: {path_with_namespace} -- we will not create it and ignore the description passed in."
        )
        assert (
            found_project.path_with_namespace == path_with_namespace
        ), f"Repository path mismatch: {found_project.path_with_namespace} != {path_with_namespace}"
        return
    except GitlabGetError:
        logger.info(f"Repository {path_with_namespace} does not exist, so we will proceed with creating it.")

    group = gl.groups.get(project_reference.subgroup)
    logger.debug(f"Found subgroup: {project_reference.subgroup}")

    # Create the repository
    project_data = {
        "name": project_reference.project_name,
        "path": project_reference.project_name,
        "namespace_id": group.id,
        "visibility": visibility,
        "initialize_with_readme": False,
        "description": description or f"Repository for {project_reference.project_name}",
    }
    logger.debug(f"Creating repository with data: {project_data}")

    created_project = gl.projects.create(project_data)
    logger.success(f"Successfully created repository: {created_project.path_with_namespace}")
    assert (
        created_project.path_with_namespace == path_with_namespace
    ), f"Repository path mismatch: {created_project.path_with_namespace} != {path_with_namespace}"


def delete_gitlab_project(
    gl: gitlab.Gitlab,
    project_reference: GitlabProjectReference,
) -> None:
    """Delete a GitLab repository.

    This operation requires "Owner" access to the subgroup.

    Args:
        gl: GitLab client instance
        project_name: Name of the repository to delete
        subgroup: Subgroup path where the repository is located

    Raises:
        GitlabGetError: If repository is not found
        GitlabDeleteError: If repository deletion fails
    """
    path_with_namespace = project_reference.path_with_namespace
    logger.info(f"Deleting GitLab repository: {path_with_namespace}")

    project = gl.projects.get(path_with_namespace)
    logger.debug(f"Found repository: {path_with_namespace}")

    # Delete the project
    project.delete()
    logger.success(f"Successfully deleted repository: {path_with_namespace}")


def is_project_already_existing(gl: gitlab.Gitlab, project_reference: GitlabProjectReference) -> bool:
    try:
        project = gl.projects.get(project_reference.path_with_namespace)
        return not bool(project.marked_for_deletion_at)
    except GitlabGetError:
        return False
    return True



"""Tests for the behaviour of errors in our system."""

from imbue_core.async_monkey_patches import log_exception
from imbue_core.errors import ImbueRuntimeException
from imbue_core.fixtures import mock_loguru_log
from imbue_core.serialization import SerializedException

_ = mock_loguru_log


def test_imbue_runtime_exceptions__logged_only_once(mock_loguru_log):
    with mock_loguru_log:
        try:
            raise ImbueRuntimeException("This is our exception")
        except ImbueRuntimeException as e:
            log_exception(e, "Ensuring  this is logged once")
            log_exception(e, "Should not be logged twice")

    assert len(mock_loguru_log.get_errors()) == 1, "This should not have logged again"


def test_imbue_runtime_exceptions__retain_logged_status(mock_loguru_log):
    with mock_loguru_log:
        try:
            raise ImbueRuntimeException("This is our exception")
        except ImbueRuntimeException as e:
            log_exception(e, "Ensuring  this is logged once")
            assert e._was_logged_by_log_exception
            sre = SerializedException.build(e)

        reconstructed_exception = sre.construct_instance()
        assert reconstructed_exception._was_logged_by_log_exception
        log_exception(reconstructed_exception, "This should not actually log")

    assert len(mock_loguru_log.get_errors()) == 1, "This should not have logged again"



from contextlib import asynccontextmanager
from pathlib import Path
from typing import AsyncIterator

import pytest

from imbue_core.common import TEMP_DIR
from imbue_core.computing_environment.computing_environment import get_untracked_files
from imbue_core.computing_environment.computing_environment import git_add
from imbue_core.computing_environment.computing_environment import is_repo_dirty
from imbue_core.computing_environment.computing_environment import make_commit
from imbue_core.conftest import mock_repo_
from imbue_core.git import LocalGitRepo
from imbue_core.git import WritableLocalGitRepo
from imbue_core.git import copy_files_from_one_repo_to_another
from imbue_core.git_snapshot import checkout_repo_from_snapshot
from imbue_core.git_snapshot import get_local_repo_snapshot
from imbue_core.testing_utils import async_temp_dir
from imbue_core.testing_utils import use


@asynccontextmanager
async def temp_writable_local_git_repo(existing_repo: LocalGitRepo) -> AsyncIterator[WritableLocalGitRepo]:
    """Context manager to create a writable version of an existing git repo.

    Under the hood this will create a copy of the existing repo in a temporary directory in the fastest way possible.
    """
    async with async_temp_dir(TEMP_DIR) as writable_repo_path:
        repo_snapshot = await get_local_repo_snapshot(existing_repo)

        cloned_repo = await checkout_repo_from_snapshot(
            repo_snapshot=repo_snapshot,
            destination_path=writable_repo_path,
        )

        # copy any untracked files into new writable repo
        if repo_snapshot.git_untracked_files is not None:
            await copy_files_from_one_repo_to_another(
                src_repo_path=existing_repo.base_path,
                dst_repo_path=writable_repo_path,
                relative_file_paths=repo_snapshot.git_untracked_files,
            )

        yield await WritableLocalGitRepo.build_from_repo(cloned_repo)


@pytest.mark.asyncio
@use(mock_repo_)
async def test_temp_writable_local_git_repo_clean(mock_repo: LocalGitRepo) -> None:
    await mock_repo.assert_clean()

    async with temp_writable_local_git_repo(mock_repo) as writable_repo:
        assert mock_repo.base_path != writable_repo.base_path
        main_repo_hash = await mock_repo.head_hash()
        writable_repo_hash = await writable_repo.head_hash()
        assert main_repo_hash == writable_repo_hash
        assert writable_repo_hash == writable_repo.initial_git_hash

        main_file_relative_path = Path("mock_repo/main.py")
        main_module_contents = await writable_repo.safely_read_file_from_repo(main_file_relative_path)

        new_code = "\n".join(
            [
                "",
                "",
                "def a_new_function(arg_1: int) -> str:",
                "    return str(arg_1)",
                "",
            ]
        )
        main_module_contents += new_code

        await writable_repo.apply_change_to_file(main_file_relative_path, main_module_contents)
        assert await is_repo_dirty(writable_repo)
        await mock_repo.assert_clean()
        new_writable_repo_hash = await make_commit(writable_repo, "making change")
        new_main_repo_hash = await mock_repo.head_hash()
        assert new_main_repo_hash == main_repo_hash
        assert new_writable_repo_hash != writable_repo_hash


@pytest.mark.asyncio
@use(mock_repo_)
async def test_temp_writable_local_git_repo_unstaged_uncommited(mock_repo: LocalGitRepo) -> None:
    # test when there is unstaged uncommited changes in main repo
    await mock_repo.assert_clean()

    mock_main_file_path = mock_repo.base_path / Path("mock_repo/main.py")
    mock_main_module_contents = await mock_repo.safely_read_file_from_repo(mock_main_file_path)

    new_code = "\n".join(
        [
            "",
            "",
            "def a_new_function(arg_1: int) -> str:",
            "    return str(arg_1)",
            "",
        ]
    )
    mock_main_module_contents += new_code
    mock_main_file_path.write_text(mock_main_module_contents)
    assert await is_repo_dirty(mock_repo)

    async with temp_writable_local_git_repo(mock_repo) as writable_repo:
        assert mock_repo.base_path != writable_repo.base_path
        await writable_repo.assert_clean()
        main_repo_hash = await mock_repo.head_hash()
        writable_repo_hash = await writable_repo.head_hash()
        # expect these to be different, since writable repo creates a new commit to manage uncommited changes
        assert main_repo_hash != writable_repo_hash
        assert main_repo_hash == writable_repo.initial_git_hash
        assert writable_repo_hash == writable_repo.stash_git_hash

        writable_main_relative_path = Path("mock_repo/main.py")
        writable_main_module_contents = await writable_repo.safely_read_file_from_repo(writable_main_relative_path)
        assert writable_main_module_contents == mock_main_module_contents


@pytest.mark.asyncio
@use(mock_repo_)
async def test_temp_writable_local_git_repo_staged_uncommited(mock_repo: LocalGitRepo) -> None:
    # test when there is staged uncommited changes in main repo
    await mock_repo.assert_clean()

    mock_main_file_path = mock_repo.base_path / Path("mock_repo/main.py")
    mock_main_module_contents = await mock_repo.safely_read_file_from_repo(mock_main_file_path)
    new_code = "\n".join(
        [
            "",
            "",
            "def a_new_function(arg_1: int) -> str:",
            "    return str(arg_1)",
            "",
        ]
    )
    mock_main_module_contents += new_code
    mock_main_file_path.write_text(mock_main_module_contents)
    assert await is_repo_dirty(mock_repo)

    before_hash = await mock_repo.head_hash()
    # stage change
    await git_add(mock_repo, str(mock_main_file_path))
    after_hash = await mock_repo.head_hash()
    assert before_hash == after_hash
    assert await is_repo_dirty(mock_repo)

    async with temp_writable_local_git_repo(mock_repo) as writable_repo:
        assert mock_repo.base_path != writable_repo.base_path
        main_repo_hash = await mock_repo.head_hash()
        assert main_repo_hash == after_hash

        await writable_repo.assert_clean()
        writable_repo_hash = await writable_repo.head_hash()
        # expect these to be different, since writable repo creates a new commit to manage uncommited changes
        assert main_repo_hash != writable_repo_hash
        assert main_repo_hash == writable_repo.initial_git_hash
        assert writable_repo_hash == writable_repo.stash_git_hash

        writable_main_relative_path = Path("mock_repo/main.py")
        writable_main_module_contents = await writable_repo.safely_read_file_from_repo(writable_main_relative_path)
        assert writable_main_module_contents == mock_main_module_contents


@pytest.mark.asyncio
@use(mock_repo_)
async def test_temp_writable_local_git_repo_untracked(mock_repo: LocalGitRepo) -> None:
    # test when there is untracked changes in main repo
    await mock_repo.assert_clean()

    new_file_path = mock_repo.base_path / "mock_repo" / "new_file.py"
    new_file_content = "\n".join(
        [
            "",
            "",
            "def hello_world(arg_1: int) -> str:",
            "    return str(arg_1)",
            "",
        ]
    )
    new_file_path.write_text(new_file_content)
    assert await is_repo_dirty(mock_repo)

    async with temp_writable_local_git_repo(mock_repo) as writable_repo:
        assert mock_repo.base_path != writable_repo.base_path
        await writable_repo.assert_clean()
        main_repo_hash = await mock_repo.head_hash()
        writable_repo_hash = await writable_repo.head_hash()
        # expect these to be different, since writable repo creates a new commit to manage uncommited changes
        assert main_repo_hash != writable_repo_hash
        assert main_repo_hash == writable_repo.initial_git_hash
        assert writable_repo_hash == writable_repo.stash_git_hash

        writable_new_relative_path = Path("mock_repo/new_file.py")
        writable_new_file_contents = await writable_repo.safely_read_file_from_repo(writable_new_relative_path)
        assert writable_new_file_contents == new_file_content


@pytest.mark.asyncio
@use(mock_repo_)
async def test_identify_untracked_files(mock_repo: LocalGitRepo) -> None:
    await mock_repo.assert_clean()

    new_file_relative_path = Path("new_file.py")
    new_file_path = mock_repo.base_path / new_file_relative_path
    new_file_content = "\n".join(
        [
            "",
            "",
            "def hello_world(arg_1: int) -> str:",
            "    return str(arg_1)",
            "",
        ]
    )
    new_file_path.write_text(new_file_content)
    untracked_files = await get_untracked_files(mock_repo)
    assert str(new_file_relative_path) in untracked_files



from pathlib import Path
from typing import Dict
from typing import Optional
from typing import cast

from imbue_core.computing_environment.computing_environment import ComputingEnvironment
from imbue_core.computing_environment.computing_environment import get_git_folder_paths
from imbue_core.computing_environment.computing_environment import get_head_hash
from imbue_core.computing_environment.computing_environment import get_modified_file_contents_by_path
from imbue_core.computing_environment.computing_environment import get_staged_unstaged_and_combined_diffs
from imbue_core.computing_environment.computing_environment import get_unmerged_and_staged_blob_contents_by_hash
from imbue_core.computing_environment.computing_environment import is_repo_conflicted
from imbue_core.frozen_utils import FrozenDict
from imbue_core.repo_state import CleanRepoOperation
from imbue_core.repo_state import ConflictType
from imbue_core.repo_state import ConflictedRepoOperation
from imbue_core.repo_state import GIT_FILE_PATH_NAMES_BY_CONFLICT_TYPE
from imbue_core.repo_state import RepoState


async def get_special_git_file_contents_by_path_for_conflict_type(
    computing_environment: ComputingEnvironment, conflict_type: ConflictType
) -> Dict[str, bytes]:
    filenames = GIT_FILE_PATH_NAMES_BY_CONFLICT_TYPE[conflict_type]
    git_file_contents_by_path: Dict[str, bytes] = {}
    for filename in filenames:
        content = await computing_environment.read_file(f".git/{filename}", mode="rb")
        assert isinstance(content, bytes), f"Expected bytes, got {type(content)}"
        git_file_contents_by_path[filename] = content
    return git_file_contents_by_path


async def get_conflict_type_from_computing_environment(
    computing_environment: ComputingEnvironment,
) -> Optional[ConflictType]:
    if await is_repo_conflicted(computing_environment):
        files = await get_git_folder_paths(computing_environment)
        if "MERGE_HEAD" in files:
            return ConflictType.MERGE
        # elif "REBASE_HEAD" in files:
        #     return ConflictType.REBASE
        # elif "CHERRY_PICK_HEAD" in files:
        #     return ConflictType.CHERRY_PICK
        else:
            return ConflictType.APPLY
    else:
        return None


async def maybe_get_conflict_operation_from_computing_environment(
    computing_environment: ComputingEnvironment,
) -> ConflictedRepoOperation | None:
    conflict_type = await get_conflict_type_from_computing_environment(computing_environment)
    if conflict_type is None:
        return None
    return await get_conflict_operation_from_computing_environment(computing_environment, conflict_type)


async def get_conflict_operation_from_computing_environment(
    computing_environment: ComputingEnvironment, conflict_type: ConflictType
) -> ConflictedRepoOperation:
    index_content = await computing_environment.read_file(Path(".git/index"), mode="rb")
    index_content = cast(bytes, index_content)
    blob_content_by_hash = await get_unmerged_and_staged_blob_contents_by_hash(computing_environment)
    modified_file_contents_by_path = await get_modified_file_contents_by_path(computing_environment)
    special_git_file_contents_by_path = await get_special_git_file_contents_by_path_for_conflict_type(
        computing_environment, conflict_type
    )
    return ConflictedRepoOperation(
        blob_content_by_hash=FrozenDict(blob_content_by_hash),
        index_content=index_content,
        modified_file_contents_by_path=FrozenDict(modified_file_contents_by_path),
        conflict_type=conflict_type,
        special_git_file_contents_by_path=FrozenDict(special_git_file_contents_by_path),
    )


async def get_clean_repo_operation_from_computing_environment(
    computing_environment: ComputingEnvironment,
) -> CleanRepoOperation:
    staged_diff, unstaged_diff, combined_diff = await get_staged_unstaged_and_combined_diffs(computing_environment)
    return CleanRepoOperation(
        combined_diff=combined_diff,
        staged_diff=staged_diff,
        unstaged_diff=unstaged_diff,
    )


async def get_conflicted_repo_state_from_computing_environment(
    computing_environment: ComputingEnvironment,
) -> RepoState:
    conflict_type = await get_conflict_type_from_computing_environment(computing_environment)
    assert conflict_type is not None
    conflict_operation = await get_conflict_operation_from_computing_environment(computing_environment, conflict_type)
    return RepoState(
        git_hash=await get_head_hash(computing_environment),
        repo_operation=conflict_operation,
    )



try:
    import pydevd_pycharm
except ImportError:
    pydevd_pycharm = None


def dbg(suspend: bool = True) -> None:
    if pydevd_pycharm is None:
        raise ImportError("pydevd_pycharm not found. Make sure you have the `pydevd-pycharm` package installed.")
    pydevd_pycharm.settrace("localhost", port=42000, stdoutToServer=True, stderrToServer=True, suspend=suspend)



import asyncio
import inspect
import os
import shutil
from contextlib import asynccontextmanager
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from typing import AsyncGenerator
from typing import Callable
from typing import ContextManager
from typing import Coroutine
from typing import Generator
from typing import Iterable
from typing import Optional
from typing import Protocol
from typing import Sequence
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union
from uuid import uuid4

import anyio
import pytest
import pytest_asyncio
from _pytest.fixtures import Config
from _pytest.python import Function

from imbue_core.common import TEMP_DIR

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from _pytest.fixtures import _ScopeName

T = TypeVar("T")

_TestFunc = Union[Callable[..., None], Callable[..., Coroutine[Any, Any, None]]]


def fixture(
    fixture_function: Optional[Any] = None,
    *,
    scope: "Union[_ScopeName, Callable[[str, Config], _ScopeName]]" = "function",
    params: Optional[Iterable[object]] = None,
    autouse: bool = False,
    ids: Optional[Union[Sequence[Optional[object]], Callable[[Any], Optional[object]]]] = None,
) -> Any:
    def decorator(function: Any) -> Any:
        true_name = function.__name__[:-1]
        if inspect.iscoroutinefunction(function):
            return pytest_asyncio.fixture(
                function, name=true_name, scope=scope, params=params, autouse=autouse, ids=ids  # type: ignore
            )
        else:
            return pytest.fixture(function, name=true_name, scope=scope, params=params, autouse=autouse, ids=ids)

    if fixture_function is not None and callable(fixture_function):
        return decorator(fixture_function)

    return decorator


def use(*args: Callable[..., Any]) -> Any:
    true_names = [x.__name__[:-1] for x in args]
    return pytest.mark.usefixtures(*true_names)


def integration_test(function: _TestFunc) -> Any:
    return pytest.mark.integration_test(function)


def slow_integration_test(function: _TestFunc) -> Any:
    return pytest.mark.slow_integration_test(function)


class RequestFixture(Protocol[T]):
    """Yes, there is a class called FixtureRequest, but the types are quite bad for it"""

    node: Function
    param: T


@fixture
def temp_file_path_() -> Generator[Path, None, None]:
    with create_temp_file_path() as output:
        yield output


@fixture
def temp_path_() -> Generator[Path, None, None]:
    with temp_dir(TEMP_DIR) as output:
        yield output


def create_temp_file_path(cleanup: bool = True) -> ContextManager[Path]:
    @contextmanager
    def context() -> Generator[Path, None, None]:
        random_id = uuid4()
        output_path = os.path.join(TEMP_DIR, str(random_id))
        try:
            yield Path(output_path)
        finally:
            if cleanup and os.path.exists(output_path):
                if os.path.isfile(output_path):
                    os.remove(output_path)
                else:
                    shutil.rmtree(output_path)

    # noinspection PyTypeChecker
    return context()


def temp_dir(base_dir: str, is_uuid_concatenated: bool = False) -> ContextManager[Path]:
    @contextmanager
    def context() -> Generator[Path, None, None]:
        random_id = uuid4()
        if is_uuid_concatenated:
            output_path = Path(base_dir.rstrip("/") + "_" + str(random_id))
        else:
            output_path = Path(base_dir) / str(random_id)
        output_path.mkdir(parents=True, exist_ok=True)
        try:
            yield output_path
        finally:
            if output_path.exists():
                try:
                    shutil.rmtree(str(output_path))
                except OSError:
                    os.unlink(str(output_path))

    # noinspection PyTypeChecker
    return context()


@asynccontextmanager
async def async_temp_dir(base_dir: str, is_uuid_concatenated: bool = False) -> AsyncGenerator[Path, None]:
    random_id = uuid4()
    if is_uuid_concatenated:
        output_path = anyio.Path(base_dir.rstrip("/") + "_" + str(random_id))
    else:
        output_path = anyio.Path(base_dir) / str(random_id)
    await output_path.mkdir(parents=True, exist_ok=True)
    try:
        yield Path(output_path)
    finally:
        if await output_path.exists():
            try:
                await asyncio.to_thread(shutil.rmtree, str(output_path))
            except OSError:
                await output_path.unlink()



from types import TracebackType
from typing import Any
from typing import Dict
from typing import Optional
from typing import Self
from typing import cast

from tblib import Traceback


class FixedTraceback(Traceback):
    """
    This class exists mostly to fix a bug in tblib where tb_lasti is not properly initialized.
    We don't care about that value, so we just set it to -1.

    While I was at it, I also fixed the types for the methods we use.
    """

    def __init__(self, tb: TracebackType) -> None:
        super().__init__(tb)
        tb_next = self
        while tb_next:
            setattr(tb_next, "tb_lasti", -1)
            tb_next = tb_next.tb_next

    def as_traceback(self) -> Optional[TracebackType]:
        return cast(Optional[TracebackType], super().as_traceback())

    @classmethod
    def from_tb(cls, tb: TracebackType) -> Self:
        result = cls(tb)
        return result

    @classmethod
    def from_dict(cls, dct: Dict[str, Any]) -> Self:
        return cast(Self, super().from_dict(dct))



# mypy: ignore-errors # TODO: @josh to fix the types in this file ;)
import asyncio
import os
from asyncio import Task
from contextvars import ContextVar
from contextvars import copy_context
from typing import Any
from typing import Coroutine
from typing import Optional
from typing import TypeVar

from loguru import logger

from imbue_core.async_monkey_patches import log_exception
from imbue_core.async_monkey_patches import safe_cancel
from imbue_core.async_monkey_patches import safe_cancel_multiple_and_wait_for_cleanup

CURRENT_TASK_GROUP_FOR_SIMPLE_EXECUTION = ContextVar("The current task group for this task")
CURRENT_TASK_GROUP_FOR_SIMPLE_EXECUTION.set(None)


T = TypeVar("T")

SET_TO_FALSE_TO_IGNORE_RAISING_DURING_TESTING_UGH_SO_EXCITED_FOR_NO_ASYNCIO = True


def create_clean_task(
    # the hammer to run
    coro: Coroutine[Any, Any, T],
    # a name to assign to the hammer Task. By default, it will be f"{function.__name__} ({hammer_id})"
    name: Optional[str] = None,
    # if true, will create a new containing task group rather than adding to the current one
    is_within_new_group: bool = False,
) -> Task:
    """
    Use this method to create a Task (rather than using asyncio.create_task directly).

    The interface is roughly the same, except that the tasks will be added to a task group for this task, if it exists,
    so that they can be properly cleaned up when the task finishes.

    The main reason you want something like this is that it's an antipattern to make async tasks without thinking
    about what group they are part of (they can be left running afterward, which usually leads to bugs).
    See this excellent blog post for more details:
    https://vorpus.org/blog/notes-on-structured-concurrency-or-go-statement-considered-harmful/
    """

    current_task_group = CURRENT_TASK_GROUP_FOR_SIMPLE_EXECUTION.get()
    if is_within_new_group:
        context = copy_context()
        new_task_group = asyncio.TaskGroup()
        # noinspection PyArgumentList
        task = asyncio.create_task(
            _run_clean_task_in_group(new_task_group, coro, name=name, context=context),
            name=f"wrapper for {name}",
            context=context,
        )
        task.owned_task_group = new_task_group
    else:
        if current_task_group is None:
            raise Exception(
                "Clean tasks must be started from another clean task, or by setting is_within_new_group to True (and being careful to actually close that task appropriately)"
            )
        else:
            task = current_task_group.create_task(coro, name=name)
            task.task_group = current_task_group
    return task


async def _run_clean_task_in_group(
    task_group: asyncio.TaskGroup, coro: Coroutine[Any, Any, T], name: str, context
) -> T:
    """We define that in generally, no exceptions will escape this context.

    Only KeyboardInterrupt and CancelledError are allowed to escape for specific reasons described below.
    """
    try:
        this_task = asyncio.current_task()
        # TODO: just for typing purposes, we may want to explicitly invoke our PropagatingTaskGroup here instead
        #  then we can make methods that call safe_cancel on all tasks in the group, etc
        async with task_group:
            task_group.name = "for " + this_task.get_name()
            this_task.task_group = task_group
            CURRENT_TASK_GROUP_FOR_SIMPLE_EXECUTION.set(task_group)
            inner_task = task_group.create_task(coro, name=name, context=context)
            inner_task.task_group = task_group
            try:
                result = await inner_task
            except asyncio.CancelledError:
                if getattr(this_task, "is_being_canceled_by_us", None):
                    # logger.debug(f"Cancelling all other tasks in this task group: {this_task.get_name()}")
                    for task in task_group._tasks:
                        safe_cancel(task)
                    # must also cancel the actual task, as this cancellation may be for this task, not that one
                    safe_cancel(inner_task)
                    # logger.debug(f"Done cancelling other tasks, returning: {this_task.get_name()}")
                    return
                raise
            # We want to ensure that the "source" of cancellation is this code.
            # Otherwise, we could add a task that raises an exception to the group
            # to cancel the rest of the group without reaching into private attributes.
            logger.debug(
                f"Cancelling all tasks in this task group because the task exited normally: {this_task.get_name()}"
            )
            to_cancel = []
            for task in task_group._tasks:
                if not task.done():
                    to_cancel.append(task)
            # FIXME: the above calls to safe_cancel() really probably need to be changed to this as well...
            #  or at least we ought to await all of the tasks (that are not our own)
            # we MUST wait here for each of the tasks to have been cancelled, otherwise we end up with some
            # pretty hard to track down bugs (which are fixed by doing this)
            # basically you can end up causing a canceled error in some lower level library, but not reaping it
            # until some other poor, totally unrelated task later comes along and tries to do something innocent
            # and then it looks like it was canceled, which means that it effectively silently disappears!
            # this whole thing is just super annoying
            await safe_cancel_multiple_and_wait_for_cleanup(to_cancel)
            logger.debug(f"Done cancelling tasks, returning: {this_task.get_name()}")
        return result
    except asyncio.CancelledError:
        # Let's let this CancelledError propagate outwards.
        raise
    except KeyboardInterrupt:
        # The user is trying to cancel the program, let us respect them by letting this propagate outwards.
        raise
    except BaseException as e:
        log_exception(e, "Clean task failed and was never retrieved")
        is_testing = "PYTEST_CURRENT_TEST" in os.environ
        if is_testing:
            if SET_TO_FALSE_TO_IGNORE_RAISING_DURING_TESTING_UGH_SO_EXCITED_FOR_NO_ASYNCIO:
                raise



"""
inlines sentry_sdk.integrations.loguru and sentry_sdk.integrations.logging, so we can make some changes.
i'm intentionally keeping most of the old logic so this still behaves roughly as expected/documented.

we probably could/should go through and fully streamline this though to do just what we need.

The changes so far (could be out of date):
- adds `strip_extra` to the breadcrumb handler
- adds `add_extra_info_hook` to the event handler, with a watchdog to make sure it doesn't slow things down
"""
import enum
import logging
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from datetime import timezone
from fnmatch import fnmatch
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Tuple

import sentry_sdk
from sentry_sdk import new_scope
from sentry_sdk.attachments import Attachment

# "This disables recording (both in breadcrumbs and as events) calls to a logger of a specific name.  Among other uses, many of our integrations
# use this to prevent their actions being recorded as breadcrumbs. Exposed to users as a way to quiet spammy loggers."
# We have to import it so that existing setters work properly
from sentry_sdk.integrations.logging import _IGNORED_LOGGERS
from sentry_sdk.types import Event
from sentry_sdk.types import Hint
from sentry_sdk.utils import current_stacktrace
from sentry_sdk.utils import event_from_exception
from sentry_sdk.utils import to_string

from imbue_core.constants import HIGH_PRIORITY_LEVEL
from imbue_core.constants import LOW_PRIORITY_LEVEL
from imbue_core.constants import MEDIUM_PRIORITY_LEVEL

# for formatting the log message. we don't want the timestamp/level because sentry already tracks that,
# and it messes up event grouping since this string becomes the event title.
SENTRY_LOG_FORMAT = "{name}:{function}:{line} - {message}"


class SentryLoguruLoggingLevels(enum.IntEnum):
    TRACE = 5
    DEBUG = 10
    INFO = 20
    SUCCESS = 25
    WARNING = 30
    # additional loguru levels for sentry hotwiring
    LOW_PRIORITY = LOW_PRIORITY_LEVEL
    MEDIUM_PRIORITY = MEDIUM_PRIORITY_LEVEL
    HIGH_PRIORITY = HIGH_PRIORITY_LEVEL
    ERROR = 40
    CRITICAL = 50


class _BaseHandler(logging.Handler):
    COMMON_RECORD_ATTRS = frozenset(
        (
            "args",
            "created",
            "exc_info",
            "exc_text",
            "filename",
            "funcName",
            "levelname",
            "levelno",
            "linenno",
            "lineno",
            "message",
            "module",
            "msecs",
            "msg",
            "name",
            "pathname",
            "process",
            "processName",
            "relativeCreated",
            "stack",
            "tags",
            "taskName",
            "thread",
            "threadName",
            "stack_info",
        )
    )

    def _can_record(self, record: logging.LogRecord) -> bool:
        """Prevents ignored loggers from recording"""
        for logger in _IGNORED_LOGGERS:
            if fnmatch(record.name, logger):
                return False
        return True

    def _extra_from_record(self, record: logging.LogRecord) -> dict[str, object]:
        return {
            k: v
            for k, v in vars(record).items()
            if k not in self.COMMON_RECORD_ATTRS and (not isinstance(k, str) or not k.startswith("_"))
        }

    def _logging_to_event_level(self, record: logging.LogRecord) -> str:
        try:
            return SentryLoguruLoggingLevels(record.levelno).name.lower()
        except ValueError:
            return record.levelname.lower() if record.levelname else ""


class SentryEventHandler(_BaseHandler):
    """A logging handler that emits Sentry events for each log record."""

    def __init__(
        self,
        level: int = logging.NOTSET,
        add_extra_info_hook: Optional[Callable[[Event, Hint], Tuple[Event, Hint]]] = None,
    ) -> None:
        super().__init__(level=level)
        self.add_extra_info_hook = add_extra_info_hook
        self.add_extra_info_previously_timed_out = False

    def emit(self, record: logging.LogRecord) -> Any:
        self.format(record)
        return self._emit(record)

    def _emit(self, record: logging.LogRecord) -> None:
        if not self._can_record(record):
            return

        client = sentry_sdk.get_client()
        if not client.is_active():
            return

        client_options = client.options

        # exc_info might be None or (None, None, None)
        #
        # exc_info may also be any falsy value due to Python stdlib being
        # liberal with what it receives and Celery's billiard being "liberal"
        # with what it sends. See
        # https://github.com/getsentry/sentry-python/issues/904
        if record.exc_info and record.exc_info[0] is not None:
            event, hint = event_from_exception(
                record.exc_info,
                client_options=client_options,
                mechanism={"type": "logging", "handled": True},
            )
        elif (record.exc_info and record.exc_info[0] is None) or record.stack_info:
            event = {}
            hint = {}
            event["threads"] = {
                "values": [
                    {
                        "stacktrace": current_stacktrace(
                            include_local_variables=client_options["include_local_variables"],
                            max_value_length=client_options["max_value_length"],
                        ),
                        "crashed": False,
                        "current": True,
                    }
                ]
            }
        else:
            event = {}
            hint = {}

        hint["log_record"] = record

        level = self._logging_to_event_level(record)
        if level in {"debug", "info", "warning", "error", "critical", "fatal"}:
            # standard levels
            event["level"] = level  # type: ignore[typeddict-item]
        elif level in {"low_prio", "med_prio", "high_prio"}:
            # artificial sentry priority induction
            match level:
                case "low_prio":
                    event["level"] = "info"
                case "med_prio":
                    event["level"] = "warning"
                case "high_prio":
                    event["level"] = "error"

        event["logger"] = record.name

        # Log records from `warnings` module as separate issues
        record_captured_from_warnings_module = record.name == "py.warnings" and record.msg == "%s"
        if record_captured_from_warnings_module:
            # use the actual message and not "%s" as the message
            # this prevents grouping all warnings under one "%s" issue
            msg = record.args[0]  # type: ignore

            event["logentry"] = {
                "message": msg,
                "params": (),
            }

        else:
            event["logentry"] = {
                "message": to_string(record.msg),
                "params": record.args,
            }

        event["extra"] = self._extra_from_record(record)

        if self.add_extra_info_hook:
            event, hint = self.add_extra_with_watchdog(event, hint, timeout=1)

        sentry_sdk.capture_event(event, hint)

    def add_extra_with_watchdog(self, event: Event, hint: Hint, timeout: float) -> Event:
        """Call the add_extra_info_hook with a watchdog so we can skip it if it's slow, and get another sentry error about that."""
        if self.add_extra_info_previously_timed_out:
            return event
        if "attachments" not in hint:
            hint["attachments"] = []
        executor = ThreadPoolExecutor()
        future = executor.submit(self.add_extra_info_hook, event, hint)
        try:
            event, hint = future.result(timeout=timeout)
            executor.shutdown()
            return event, hint
        except TimeoutError as e:
            # this is unexpected; log another sentry event about it.
            log_error_inside_sentry(e, f"add_extra_info_hook took longer than {timeout} seconds")
            # this will leave the thread still running; there's no real way to cancel it.
            # we'll at least set this flag so future errors don't try to run the (bugged?) hook again.
            executor.shutdown(wait=False)
            self.add_extra_info_previously_timed_out = True

        # continue with the main event without the extra info
        return event, hint


class SentryBreadcrumbHandler(_BaseHandler):
    """
    A logging handler that records breadcrumbs for each log record.

    Note that you do not have to use this class if the logging integration is enabled, which it is by default.
    """

    def __init__(self, level: int = logging.NOTSET, strip_extra: bool = False) -> None:
        super().__init__(level=level)
        self.strip_extra = strip_extra

    def emit(self, record: logging.LogRecord) -> Any:
        self.format(record)
        return self._emit(record)

    def _emit(self, record: logging.LogRecord) -> None:
        if not self._can_record(record):
            return

        sentry_sdk.add_breadcrumb(self._breadcrumb_from_record(record), hint={"log_record": record})

    def _breadcrumb_from_record(self, record: logging.LogRecord) -> Dict[str, Any]:
        return {
            "type": "log",
            "level": self._logging_to_event_level(record),
            "category": record.name,
            "message": record.message,
            "timestamp": datetime.fromtimestamp(record.created, timezone.utc),
            "data": self._extra_from_record(record) if not self.strip_extra else {},
        }


def log_error_inside_sentry(
    exception: Exception,
    message: str,
    extra: Optional[Dict[str, str]] = None,
    attachments: Optional[Iterable[Attachment]] = None,
) -> None:
    """Log an error to sentry that happens during processing of a sentry event.

    This needs to be done very carefully to ensure it won't fail - we don't want to have to have a fallback-fallback handler.
    The caller should ensure everything passed into this is small so there's no chance of size issues.
    """
    client = sentry_sdk.get_client()
    # we want to get rid of any breadcrumbs, attachments, and other stuff that might have caused the original request to fail.
    # this will obviously make it harder to debug; we may want to selectively add some of this back.
    with new_scope() as scope:
        scope.clear()
        event, hint = event_from_exception(
            exception, client_options=client.options, mechanism={"type": "watchdog", "handled": True}
        )
        event["message"] = message
        if extra is not None:
            if "extra" not in event:
                event["extra"] = {}
            for k, v in extra.items():
                event["extra"][k] = v
        if attachments is not None:
            if "attachments" not in hint:
                hint["attachments"] = []
            for attachment in attachments:
                hint["attachments"].append(attachment)
        # Note that new_scope() gives a new "current scope" but doesn't affect the global or isolation scope,
        # which is where most info is actually stored. Typically all 3 scopes are merged before logging the event.
        # So we'll make sure to call capture_event in such a way that this merging doesn't happen.
        client.capture_event(event=event, hint=hint, scope=scope)



import functools
import os
import sys
import traceback
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type

import attr
import grpclib.exceptions
import sentry_sdk
import sentry_sdk.utils
import traceback_with_variables
from loguru import logger
from sentry_sdk import HttpTransport
from sentry_sdk.attachments import Attachment
from sentry_sdk.consts import EndpointType
from sentry_sdk.envelope import Envelope
from sentry_sdk.integrations.stdlib import StdlibIntegration
from sentry_sdk.types import Event
from sentry_sdk.types import Hint
from traceback_with_variables import Format

from imbue_core.common import truncate_string
from imbue_core.sentry_loguru_handler import SENTRY_LOG_FORMAT
from imbue_core.sentry_loguru_handler import SentryBreadcrumbHandler
from imbue_core.sentry_loguru_handler import SentryEventHandler
from imbue_core.sentry_loguru_handler import SentryLoguruLoggingLevels
from imbue_core.sentry_loguru_handler import log_error_inside_sentry

try:
    import brotli  # type: ignore
except ImportError:
    brotli = None


# sentry's size limits are annoyingly hard to evaluate before sending the event. we'll just try to be conservative.
# https://docs.sentry.io/concepts/data-management/size-limits/
# https://develop.sentry.dev/sdk/data-model/envelopes/#size-limits
MAX_SENTRY_ATTACHMENT_SIZE = 10 * 1024 * 1024

RATE_LIMITED_EXCEPTION_TYPES = (grpclib.exceptions.GRPCError,)


class SentryEventRejected(Exception):
    pass


ExceptionKeyType = Tuple[Type[BaseException], Tuple[Any, ...]]


def _create_key_from_exception(exception: BaseException) -> ExceptionKeyType:
    return type(exception), exception.args


# TODO: we could introduce another more sophisticated rate limiting mechanism here that checks the frequency of an event and rate limits it automatically
@attr.s(auto_attribs=True, kw_only=True)
class _ManualSentryEventRateLimiter:
    """Prevent logging the same specific exceptions multiple times to sentry."""

    _seen_exceptions: Set[ExceptionKeyType] = attr.ib(factory=set)

    def _record_exception(self, exception: BaseException) -> None:
        self._seen_exceptions.add(_create_key_from_exception(exception))

    def __len__(self) -> int:
        return len(self._seen_exceptions)

    def __contains__(self, exception: BaseException) -> bool:
        return _create_key_from_exception(exception) in self._seen_exceptions

    def is_worth_logging_exception(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        rate_limited_exceptions: Iterable[Type[BaseException]] = RATE_LIMITED_EXCEPTION_TYPES,
    ) -> bool:
        if exc_type in rate_limited_exceptions:
            if exc_value in self:
                logger.trace("Already logged the following exception to sentry, skipping: {}", exc_value)
                return False
            logger.trace("Logging the following exception to sentry and rate limiting future instances: {}", exc_value)
            self._record_exception(exc_value)
        return True


class ImbueSentryHttpTransport(HttpTransport):
    """The sentry python sdk has pretty lame behavior if the event is too large.
    It'll just drop it, and record stats indicating that an event was dropped.
    You can see these at `https://generally-intelligent-e3.sentry.io/stats`, category "invalid".
    But there's no way to recover any information about the dropped event.

    We could try to just ensure the events don't violate the size limit, which we try to do,
    but their size limits are a bit complicated and thus hard to pre-verify. So we also want to know if anything slips through.

    The actual sentry web API does return a status code (413) if the event was rejected,
    so we need to handle this at the level of the sentry HttpTransport and do something with it.
    """

    def _send_request(
        self,
        body: bytes,
        headers: Dict[str, str],
        endpoint_type: EndpointType = EndpointType.ENVELOPE,
        envelope: Optional[Envelope] = None,
    ) -> None:
        """This is a copy of the original `_send_request` method from the HttpTransport class,
        with a hook to call `on_too_large_event` added.
        """

        def record_loss(reason: str) -> None:
            if envelope is None:
                self.record_lost_event(reason, data_category="error")
            else:
                for item in envelope.items:
                    self.record_lost_event(reason, item=item)

        headers.update(
            {
                "User-Agent": str(self._auth.client),
                "X-Sentry-Auth": str(self._auth.to_header()),
            }
        )
        try:
            response = self._request(
                "POST",
                endpoint_type,
                body,
                headers,
            )
        except Exception:
            self.on_dropped_event("network")
            record_loss("network_error")
            raise

        try:
            self._update_rate_limits(response)

            if response.status == 429:
                # if we hit a 429.  Something was rate limited but we already
                # acted on this in `self._update_rate_limits`.  Note that we
                # do not want to record event loss here as we will have recorded
                # an outcome in relay already.
                self.on_dropped_event("status_429")

            elif response.status >= 300 or response.status < 200:
                sentry_sdk.utils.logger.error(
                    "Unexpected status code: %s (body: %s)",
                    response.status,
                    getattr(response, "data", getattr(response, "content", None)),
                )
                self.on_dropped_event("status_{}".format(response.status))
                record_loss("network_error")

                if response.status == 413:
                    self.on_too_large_event(body, envelope)
        finally:
            response.close()

    def on_too_large_event(self, body: bytes, envelope: Envelope) -> None:
        """we want to log _something_ to sentry, because otherwise we have no idea what happened,
        but we also need to be super careful that this fallback doesn't itself fail.

        exceptions raised here will simply get eaten and result in nothing getting logged to sentry,
        both due to sentry's usage of `capture_internal_exceptions`
        and that we're running in a worker thread and i don't think they make an effort to re-surface exceptions from threads.
        """
        msg = f"request was too large to send to sentry"
        try:
            raise SentryEventRejected(msg)
        except SentryEventRejected as e:
            stripped_envelope = Envelope(headers=envelope.headers)
            attachment_sizes = {}
            for item in envelope.items:
                if item.data_category == "attachment":
                    attachment_sizes[item.headers["filename"]] = len(item.payload.get_bytes())
                    continue
                stripped_envelope.add_item(item)
            # this is uncompressed (so we can inspect it)
            serialized_stripped_envelope = stripped_envelope.serialize()

            extra = {
                "uncompressed_attachment_sizes": str(attachment_sizes),
                "original_compressed_request_body_size": len(body),
                "uncompressed_stripped_envelope_size": len(serialized_stripped_envelope),
            }

            # attach stripped envelope as attachment. ensure we're well under the attachment size limit.
            if len(serialized_stripped_envelope) >= MAX_SENTRY_ATTACHMENT_SIZE:
                serialized_stripped_envelope = serialized_stripped_envelope[: MAX_SENTRY_ATTACHMENT_SIZE - 3] + b"..."
            attachment = Attachment(serialized_stripped_envelope, filename="stripped_envelope.txt")
            log_error_inside_sentry(e, msg, extra=extra, attachments=(attachment,))


def get_traceback_with_vars(exception: Optional[BaseException] = None) -> str:
    # be careful of potential performance regressions with increasing these limits
    tb_format = Format(max_value_str_len=100_000, max_exc_str_len=2_000_000)
    if exception is None:
        # no exception passed in; get the current exception. this will still be None if not in an exception handler
        exception = sys.exception()
    try:
        if exception is not None:
            # we are in an exception handler, use that for the traceback
            # for some reason this breaks when casting to an `Exception`, so just using type: ignore
            return traceback_with_variables.format_exc(exception, fmt=tb_format)  # type: ignore
        else:
            # not in an exception handler, just get the current stack
            return traceback_with_variables.format_cur_tb(fmt=tb_format)
    except Exception as e:
        return f"got exception while formatting traceback with `traceback_with_variables`: {traceback.format_exception(e)}"


# call get_traceback_with_vars with a timeout. yes this duplicates `SentryEventHandler.add_extra_with_watchdog`, sorry
# FIXME this could be generalized as a decorator, but the typing is difficult
def timeout_get_traceback_with_vars(exc: Optional[BaseException] = None, timeout: float = 1.0) -> Optional[str]:
    executor = ThreadPoolExecutor()
    future = executor.submit(get_traceback_with_vars, exc)
    try:
        tb_string = future.result(timeout=timeout)
        executor.shutdown()
        return tb_string
    except TimeoutError as e:
        # this is unexpected; log another sentry event about it.
        log_error_inside_sentry(e, f"get_traceback_with_vars took longer than {timeout} seconds")
        # this will leave the thread still running; there's no real way to cancel it.
        executor.shutdown(wait=False)
        return None


def _default_sentry_before_send_hook(event: Event, hint: Hint) -> Optional[Event]:
    """
    Add traceback with variables to the event as an attachment.

    """

    expected_attachments = []
    tb_with_vars = truncate_string(get_traceback_with_vars(), MAX_SENTRY_ATTACHMENT_SIZE)
    hint["attachments"].append(Attachment(tb_with_vars.encode(), filename="traceback_with_variables.txt"))
    expected_attachments.append("traceback_with_variables.txt")
    # record the names of the expected attachments just in case there's any weirdness about attachments not showing up
    event["extra"]["expected_attachments"] = str(expected_attachments)
    return event


# TODO: if the actual event (without attachments) being too large is the problem, we could use something like this
#  to handle that case and truncate whatever is being so large inside of it.
#  this would be added as a `before_send` hook in the sentry_sdk init.
# def before_send(event, hint) -> None:
#     # limit here is 1MB uncompressed or 200KB compressed.
#     # to not have to deal with computing compressed sizes, i'll just limit to 200KB uncompressed for now.
#     MAX_SENTRY_EVENT_SIZE = 200 * 1024
#     event_bytes = sentry_sdk.utils.json_dumps(event)
#     event_size = len(event_bytes)
#     if event_size > MAX_SENTRY_EVENT_SIZE:
#         # TODO: add logic here to strip out some of the event data to make it smaller
#         pass
def _before_send_hook(
    event: Event,
    hint: Hint,
    rate_limiter: _ManualSentryEventRateLimiter,
    before_send: Optional[Callable[[Event, Hint], Event]] = None,
) -> Optional[Event]:
    exc_info = hint["exc_info"]  # see from sentry_sdk._types import ExcInfo which sadly you can't import
    exc_type, exc_value, exc_tb = exc_info

    if not rate_limiter.is_worth_logging_exception(exc_type=exc_type, exc_value=exc_value):
        return None

    if before_send is not None:
        event = before_send(event, hint)
    return event


def setup_sentry(
    dsn: str,
    username: Optional[str],
    release_id: str,
    integrations: Tuple[Any, ...] = (),
    add_extra_info_hook: Optional[Callable[[Event, Hint], Tuple[Event, Hint]]] = None,
) -> None:
    """
    This should be done *after* setting up normal loguru loggers, to ensure that sentry handling happens after normal logging.
    In case the sentry stuff hangs or something odd, we want to make sure to at least get regular log output.
    """
    assert (
        "SENTRY_DSN" not in os.environ
    ), "Please `unset SENTRY_DSN` in your environment. Set the DSN via the server settings FRONTEND_SENTRY_DSN and BACKEND_SENTRY_DSN instead."

    # here we create a simple wrapper around passed in `before_send` hooks so we can always rate limit specific exceptions across all usages of `setup_sentry`
    rate_limiter = _ManualSentryEventRateLimiter()
    before_send = functools.partial(
        _before_send_hook,
        rate_limiter=rate_limiter,
        before_send=_default_sentry_before_send_hook if add_extra_info_hook is None else None,
    )

    sentry_sdk.init(
        sample_rate=1.0,
        traces_sample_rate=1.0,
        # required for `logger.error` calls to include stacktraces
        attach_stacktrace=True,
        # note this will capture unhandled exceptions even if not explicitly logged, among other things
        # https://docs.sentry.io/platforms/python/integrations/default-integrations/
        default_integrations=True,
        # this doesn't affect the default integrations, but prevents any other ones from being added automatically
        auto_enabling_integrations=False,
        integrations=[
            *integrations,
        ],
        disabled_integrations=[
            # this only adds hooks to subprocess and httplib, which imo just adds noisy breadcrumbs.
            StdlibIntegration()
        ],
        dsn=dsn,
        # may want to get more restrictive about this in the future
        # see https://docs.sentry.io/platforms/python/data-management/data-collected/
        send_default_pii=True,
        # sentry has a max payload size of 1MB, so we can't make this infinite
        max_value_length=100_000,
        add_full_stack=True,
        # If add_extra_info_hook exists, we assume it supersedes the default sentry before_send hook.
        before_send=before_send,
        release=release_id,
        # default is 100; can't make it too large because total event size must be <1MB
        max_breadcrumbs=1000,
        # if the locals is very large, sentry gets to be quite slow to log errors if this is enabled.
        # we log our own traceback_with_variables anyways.
        include_local_variables=False,
        transport=ImbueSentryHttpTransport,
    )
    logger.info("Sentry initialized")

    # capture loguru errors/exceptions with a custom handler
    min_sentry_level: int = SentryLoguruLoggingLevels.LOW_PRIORITY.value
    logger.add(
        SentryEventHandler(level=min_sentry_level, add_extra_info_hook=add_extra_info_hook),
        level=min_sentry_level,
        diagnose=False,
        format=SENTRY_LOG_FORMAT,
    )
    # capture lower level loguru messages to add as breadcrumbs on events
    # the extra info is not helpful here and makes the breadcrumbs larger; they're still available in the log file attachment
    breadcrumb_level: int = SentryLoguruLoggingLevels.INFO.value
    logger.add(
        SentryBreadcrumbHandler(level=breadcrumb_level, strip_extra=True),
        level=breadcrumb_level,
        diagnose=False,
        format=SENTRY_LOG_FORMAT,
    )

    if username is not None:
        sentry_sdk.set_user({"username": username})



#  Related comment from Sam: "One of the reasons I don't really like using subclasses for things like this is that __init_subclass__ gets called before dataclass decorators are applied (so there's not really a great way to make sure that a class that inherits from Serializable is actually serializable). I think it's fine just as a marker for now, but might be worth thinking about at some point."
class Serializable:
    pass



import asyncio
import functools
import inspect
import os
import platform
import sys
import threading
import traceback
from contextlib import AbstractAsyncContextManager
from contextlib import AbstractContextManager
from contextlib import contextmanager
from datetime import datetime
from http.server import BaseHTTPRequestHandler
from http.server import HTTPServer
from pathlib import Path
from types import FrameType
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Coroutine
from typing import Dict
from typing import Generator
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import ParamSpec
from typing import TypeVar
from typing import Union
from typing import cast
from urllib.parse import parse_qs
from urllib.parse import urlparse

from loguru import logger
from traceback_with_variables.core import _iter_lines

from imbue_core.async_monkey_patches import log_exception
from imbue_core.async_monkey_patches import safe_cancel

P = ParamSpec("P")
R = TypeVar("R")

ALL_EVENT_LOOPS: list[asyncio.AbstractEventLoop] = []


def sync(func: Callable[P, Awaitable[R]]) -> Callable[P, R]:
    """Decorator that runs an async function synchronously by dispatching to
    an event loop running in a separate thread.
    """

    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        loop = _get_or_create_event_loop()
        return asyncio.run_coroutine_threadsafe(func(*args, **kwargs), loop).result()

    return wrapper


S = TypeVar("S")


@contextmanager
def sync_contextmanager(async_context_manager: AbstractAsyncContextManager[S]) -> Generator[S, None, None]:
    sync_aenter = sync(async_context_manager.__aenter__)
    sync_aexit = sync(async_context_manager.__aexit__)

    enter_result = sync_aenter()
    try:
        yield enter_result
    except BaseException as e:
        if not sync_aexit(e.__class__, e, e.__traceback__):
            raise
    else:
        sync_aexit(None, None, None)


def sync_contextmanager_func(
    cm_func: Callable[P, AbstractAsyncContextManager[S]]
) -> Callable[P, AbstractContextManager[S]]:
    @functools.wraps(cm_func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> AbstractContextManager[S]:
        return sync_contextmanager(cm_func(*args, **kwargs))

    return wrapper


_LOOP: Optional[asyncio.AbstractEventLoop] = None
_LOOP_LOCK: threading.Lock = threading.Lock()


def _get_or_create_event_loop() -> asyncio.AbstractEventLoop:
    global _LOOP
    if _LOOP is not None:
        return _LOOP
    with _LOOP_LOCK:
        # Check again in case another thread created the loop while we were waiting for the lock.
        if _LOOP is not None:
            return _LOOP
        _LOOP = asyncio.new_event_loop()
        asyncio.set_event_loop(_LOOP)
        threading.Thread(target=_LOOP.run_forever, daemon=True, name="async_loop").start()
    return _LOOP


def shorten_filename(filename: str) -> str:
    path = Path(filename)
    while path.parent:
        path = path.parent
        if not (path / "__init__.py").exists():
            break

    try:
        shortened = str(Path(filename).relative_to(path))
    except ValueError:
        shortened = filename  # in case the path cannot be made relative

    return shortened


# TODO: I'd really like to print these task groups in a hierarchical way instead of flat -- we know which groups
#  launched which other groups, so we could print them in a tree structure. That would be a lot more readable.
#  It might also be nice to print without any stacks at all. As long as the tasks had good names, that would make it
#  possible to very easily understand everything that was currently executing.
#  I could even imagine controls that allowed for printing just the task groups themselves, which would also be easier
#  to understand.
def get_all_async_task_stacks(
    num_skipped_frames: int = 0, log_variables: bool = False, loop: Optional[asyncio.AbstractEventLoop] = None
) -> Iterator[str]:
    """Yields the lines of a report for all stack frames of all async tasks including variables."""
    tasks_by_task_group: Dict[Optional[asyncio.TaskGroup], List[asyncio.Task]] = {}
    owning_task_by_task_group: Dict[asyncio.TaskGroup, asyncio.Task] = {}

    for task in asyncio.all_tasks(loop=loop):
        if task.done():
            continue
        task_group = cast(Optional[asyncio.TaskGroup], getattr(task, "task_group", None))
        owned_task_group = cast(Optional[asyncio.TaskGroup], getattr(task, "owned_task_group", None))
        if owned_task_group is not None:
            owning_task_by_task_group[owned_task_group] = task
            if owned_task_group not in tasks_by_task_group:
                tasks_by_task_group[owned_task_group] = []
        else:
            tasks_by_task_group.setdefault(task_group, []).append(task)

    all_owning_tasks = set(owning_task_by_task_group.values())

    task_group_keys = list(tasks_by_task_group.keys())
    for task_group in cast(List[Optional[asyncio.TaskGroup]], [None]) + [x for x in task_group_keys if x is not None]:
        if task_group not in tasks_by_task_group:
            continue
        tasks = tasks_by_task_group[task_group]
        if task_group is None:
            yield f"\n\n{'='*40}\nNo TaskGroup:\n"
        else:
            yield f"\n\n{'='*40}\nTaskGroup: {getattr(task_group, 'name', 'unknown')}\n"
        owning_task = None
        if task_group is not None:
            owning_task = owning_task_by_task_group.get(task_group)
        is_first_line_skipped = False
        all_tasks = tasks
        if owning_task is not None:
            yield f"Owning Task: {owning_task.get_name()}\n"
            is_first_line_skipped = True
            all_tasks.insert(0, owning_task)
        for task in all_tasks:
            # skip these -- they'll be printed at the top of the group that they own
            if task_group is None and task in all_owning_tasks:
                continue
            if is_first_line_skipped:
                is_first_line_skipped = False
            else:
                yield f"{'-'*40}\nTask {task.get_name()}:\n"
            frames = extract_frames(task)
            for frame in frames:
                frame_infos = inspect.getouterframes(frame)[num_skipped_frames:]
                # Use private method _iter_lines to traceback async tasks, which is not explicitly handled in the API
                if log_variables:
                    for line in _iter_lines(
                        e=None,
                        frame_infos=frame_infos,
                        fmt=None,
                        for_file=None,
                    ):
                        yield line + "\n"
                else:
                    frame_summaries = [
                        traceback.FrameSummary(
                            shorten_filename(info.filename),
                            lineno=info.lineno,
                            name=info.function,
                            line=info.code_context[0].strip() if info.code_context else None,
                        )
                        for info in frame_infos
                    ]
                    yield from traceback.format_list(frame_summaries)


def extract_frames(task: asyncio.Task) -> list[FrameType]:
    """Extract the stack frames of an async task."""
    coro = task.get_coro()
    assert isinstance(coro, Coroutine)
    frames = []
    while coro is not None and coro.cr_frame is not None:
        frames.append(coro.cr_frame)
        coro = coro.cr_await  # type: ignore
        # this happens at the very bottom of the call stack, there it seems to often be a FutureIter, Event, etc
        if type(coro).__name__ != "coroutine":
            break
    return frames


def print_all_async_task_stacks(log_variables: bool = False) -> None:
    """Prints the stack frames of all running tasks."""
    for line in get_all_async_task_stacks(log_variables=log_variables):
        print(line)


def dump_all_async_task_stacks(log_path: Union[str, Path], log_variables: bool = False) -> None:
    """Dump the stack frames of all running tasks to file."""
    with open(log_path, "w") as f:
        for line in get_all_async_task_stacks(log_variables=log_variables):
            if log_variables:
                line += "\n"
            f.write(line)


async def periodically_log_async_stacks(
    log_dir: Union[str, Path], interval: float, log_variables: bool = False
) -> None:
    """Periodically print the stack traces of all running tasks."""
    Path(log_dir).mkdir(parents=True, exist_ok=True)
    while True:
        log_path = Path(log_dir) / f"{datetime.now().strftime('%Y%m%d_%H%M%S')}.log"
        dump_all_async_task_stacks(log_path=log_path, log_variables=log_variables)
        logger.debug(f"Dumped asyncio stack traces to {log_path}")
        await asyncio.sleep(interval)


async def is_task_group_complete(
    task_group: asyncio.TaskGroup, trace_task: asyncio.Task, buffer_time: float = 1.0
) -> None:
    """Continuously check if all tasks except the stack trace logger are done."""
    while True:
        if all(task.done() for task in task_group._tasks if task is not trace_task):
            await asyncio.sleep(buffer_time)  # Wait for buffer time in case new tasks are added
            # Recheck to confirm
            if all(task.done() for task in task_group._tasks if task is not trace_task):
                break
        await asyncio.sleep(1.0)


async def inject_async_stack_trace_logger(
    task_group: asyncio.TaskGroup, log_dir: Union[str, Path], log_interval: float = 60.0, log_variables: bool = False
) -> None:
    """Inject a stack trace logger into the task group."""
    trace_task = asyncio.create_task(
        periodically_log_async_stacks(log_dir=log_dir, interval=log_interval, log_variables=log_variables),
        name="periodically_log_async_stacks",
    )
    await is_task_group_complete(task_group, trace_task)
    safe_cancel(trace_task)
    try:
        await trace_task
    except asyncio.CancelledError:
        pass


class AsyncTaskStacksHandler(BaseHTTPRequestHandler):
    def do_GET(self) -> None:
        try:
            parsed_url = urlparse(self.path)
            query_params = parse_qs(parsed_url.query)
            log_variables = query_params.get("locals", ["false"])[0].lower() in ["true", "1", "yes"]

            self.send_response(200)
            self.send_header("Content-type", "text/plain")
            self.end_headers()

            pid = os.getpid()
            command_line = " ".join(sys.argv)

            # Path to the Python executable
            python_executable = sys.executable

            # Python version
            python_version = platform.python_version()

            # Print the collected information
            self.wfile.write(f"Process {pid}: {python_executable} {command_line}\n".encode("utf-8"))
            self.wfile.write(f"Python v{python_version} ({python_executable})\n\n".encode("utf-8"))

            for loop in ALL_EVENT_LOOPS:
                for line in get_all_async_task_stacks(log_variables=log_variables, loop=loop):
                    self.wfile.write(line.encode("utf-8"))
        except BaseException as e:
            log_exception(e, "exception in AsyncTaskStacksHandler")
            raise


def run_async_stackframe_server_thread(port_range_low: int, port_range_high: int) -> None:
    try:
        success = False
        for port in range(port_range_low, port_range_high):
            try:
                server_address = ("localhost", port)
                httpd = HTTPServer(server_address, AsyncTaskStacksHandler)
                success = True
                print(f"Starting async stack trace server on port {port}. Process pid: {os.getpid()}")
                break
            except OSError:
                continue

        if not success:
            logger.info("Could not find an open port to start the async stack trace server, continuing without it.")
            return

        httpd.serve_forever()
    except BaseException as e:
        log_exception(e, "exception in run_async_stackframe_server_thread")
        raise


IS_STACKFRAME_SERVER_RUNNING = False
STACKFRAME_SERVER_PORT_LOW = 60000
STACKFRAME_SERVER_PORT_HIGH = 61000


def run_async_stackframe_server_for_loop(event_loop: asyncio.AbstractEventLoop) -> None:
    ALL_EVENT_LOOPS.append(event_loop)
    run_async_stackframe_server()


def run_async_stackframe_server() -> None:
    global IS_STACKFRAME_SERVER_RUNNING
    if not IS_STACKFRAME_SERVER_RUNNING:
        IS_STACKFRAME_SERVER_RUNNING = True

        port = os.environ.get("ASYNC_STACKFRAME_SERVER_PORT")
        if port is not None:
            try:
                port = int(port)
            except ValueError:
                logger.error(f"ASYNC_STACKFRAME_SERVER_PORT is not an integer: {port}")
                raise

        if port is None:
            port_range_low = STACKFRAME_SERVER_PORT_LOW
            port_range_high = STACKFRAME_SERVER_PORT_HIGH
        else:
            port_range_low = port
            port_range_high = port + 1

        thread = threading.Thread(
            target=run_async_stackframe_server_thread,
            daemon=True,
            kwargs={"port_range_low": port_range_low, "port_range_high": port_range_high},
        )
        thread.start()


def with_timeout(func: Callable[P, Awaitable[R]], timeout_secs: float) -> Callable[P, Awaitable[R]]:
    """Decorator that adds a timeout to an async function."""

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        return await asyncio.wait_for(func(*args, **kwargs), timeout_secs)

    return wrapper


T = TypeVar("T")


async def gather_with_limited_concurrency(coros: Iterable[Awaitable[T]], n: int) -> List[T]:
    """Like asyncio.gather() but will only run `n` in parallel at a time.

    Note that a call like `asyncio.gather(*coros)` is now `gather_with_limited_concurrency(coros, n=10),
    without the splat.
    """
    semaphore = asyncio.Semaphore(n)

    async def sem_coro(coro: Awaitable[T]) -> T:
        async with semaphore:
            return await coro

    return await asyncio.gather(*(sem_coro(c) for c in coros))


_NOT_FOUND = object()


class AsyncCachedProperty(Generic[T]):
    """A descriptor factory that behaves very similarly to `functools.cached_property`, but for
    async methods!

    The type annotations here are rough; it's not realistic to get them perfect without using a .pyi file.
    """

    def __init__(self, func: Callable[[Any], Coroutine[None, None, T]]) -> None:
        self.func = func
        self.attrname: Optional[str] = None
        self.__doc__ = func.__doc__

    def __set_name__(self, owner: type, name: str) -> None:
        if self.attrname is None:
            self.attrname = name
        elif name != self.attrname:
            raise TypeError("Cannot assign the same AsyncCachedProperty to multiple names")

    def _get_attrname(self) -> str:
        if self.attrname is None:
            raise TypeError("Cannot use AsyncCachedProperty instance without calling __set_name__")
        return self.attrname

    def _get_cache(self, instance: object) -> Dict[str, Any]:
        try:
            return instance.__dict__
        except AttributeError:
            raise TypeError(
                "Cannot use AsyncCachedProperty with instances that do not have a __dict__ attribute"
            ) from None

    def __get__(self, instance: object, owner: Optional[type] = None) -> Awaitable[T]:
        if instance is None:
            return self  # type: ignore
        attrname = self._get_attrname()
        cache = self._get_cache(instance)
        val = cache.get(attrname, _NOT_FOUND)
        if val is not _NOT_FOUND:
            return cast(Awaitable[T], val)

        task = asyncio.create_task(self.func(instance))
        cache[attrname] = task
        return task

    def __delete__(self, instance: object) -> None:
        if instance is None:
            raise TypeError("Cannot delete AsyncCachedProperty on a class")
        attrname = self._get_attrname()
        cache = self._get_cache(instance)
        try:
            awaitable = cache.pop(attrname)
            if not awaitable.done():
                safe_cancel(awaitable)
        except KeyError:
            raise AttributeError(f"Cannot delete attribute {self.attrname!r}") from None

    def __set__(self, instance: object, value: T) -> None:
        if instance is None:
            raise TypeError("Cannot set AsyncCachedProperty on a class")
        attrname = self._get_attrname()
        cache = self._get_cache(instance)
        existing = cache.pop(attrname, None)
        if existing is not None and not existing.done():
            safe_cancel(existing)
        fut: asyncio.Future[T] = asyncio.Future()
        fut.set_result(value)
        cache[attrname] = fut


def wrapped_asyncio_run(
    main: Coroutine,
    *,
    debug: Optional[bool] = None,
    loop_factory: Optional[Callable[..., asyncio.AbstractEventLoop]] = None,
) -> Any:
    """
    This is a lightweight wrapper with a singular purpose -- it is here to enable apyspy

    apyspy is our async equivalent to py-spy.

    Without apyspy, it's really annoying to debug why some async task is stuck.
    """

    async def wrapper_main() -> Any:
        ALL_EVENT_LOOPS.append(asyncio.get_event_loop())
        result = await main
        ALL_EVENT_LOOPS.pop()
        return result

    run_async_stackframe_server()
    with asyncio.Runner(debug=debug, loop_factory=loop_factory) as runner:
        return runner.run(wrapper_main())


def make_async(func: Callable[P, R]) -> Callable[P, Awaitable[R]]:
    """
    Turn the annotated function into an async function by running it in a thread.

    This is useful for functions that perform blocking i/o.
    """

    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        return await asyncio.to_thread(func, *args, **kwargs)

    return wrapper



"""
Data types and other utils shared by the local sync server (backend) and client (frontend).

"""
import datetime
from enum import StrEnum
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

import attr

from imbue_core.common import generate_id
from imbue_core.repo_state import CleanRepoOperation
from imbue_core.repo_state import ConflictedRepoOperation
from imbue_core.serialization_types import Serializable
from imbue_core.time_utils import get_current_time


def _convert_to_repo_operation(v: Any) -> Union[CleanRepoOperation, ConflictedRepoOperation]:
    """Convert a dict to a repo operation."""
    if isinstance(v, (CleanRepoOperation, ConflictedRepoOperation)):
        return v
    if not isinstance(v, dict):
        raise ValueError("repo_operation must be a dict")
    for operation_class_type in (CleanRepoOperation, ConflictedRepoOperation):
        try:
            return operation_class_type(**v)
        except Exception:
            pass
    raise ValueError("repo_operation must be a CleanRepoOperation or ConflictedRepoOperation")


def _convert_to_utc_datetime(v: Any) -> datetime.datetime:
    if isinstance(v, str):
        return datetime.datetime.fromisoformat(v)
    if isinstance(v, dict):
        if "time" in v:
            # Got something like {"time": 1741999454.609277, "tzaware": true}
            return datetime.datetime.fromtimestamp(v["time"], datetime.timezone.utc)
        raise ValueError("created_at must be a string or a dict with a 'time' key")
    return v


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class Project(Serializable):
    backend_repo_url: str
    name: str
    id: str


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class LocalSyncClientInfo(Serializable):
    """Information about the client sent to the backend."""

    ip_address: str


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class LocalSyncMessage(Serializable):
    """Base class for all local sync messages."""

    created_at: datetime.datetime = attr.ib(factory=get_current_time, converter=_convert_to_utc_datetime)
    message_id: str = attr.ib(factory=generate_id)


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class LocalSyncClientMessage(LocalSyncMessage):
    """Message from the client to send to the backend."""

    client_id: str
    client_info: LocalSyncClientInfo


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class CreateProjectMessage(LocalSyncClientMessage):
    project_name: str


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class UploadLogsMessage(LocalSyncClientMessage):
    log_entries: Tuple[str, ...]


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class SyncBranchMessage(LocalSyncClientMessage):
    project_id: str
    branch_name: str
    target_branch_name: str
    commit_hash: str
    repo_operation: Union[CleanRepoOperation, ConflictedRepoOperation] = attr.ib(converter=_convert_to_repo_operation)
    user_choice: Optional[str] = None


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class NewLocalRepoStateMessage(LocalSyncClientMessage):
    """Serialized message from the backend to send to the client."""

    project_id: str
    branch_name: str
    version: int
    commit_hash: str
    repo_operation: Union[CleanRepoOperation, ConflictedRepoOperation] = attr.ib(converter=_convert_to_repo_operation)


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class NewSecretsMessage(LocalSyncClientMessage):
    encrypted_secrets_file_contents: bytes


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class NewSecretKeyMessage(LocalSyncClientMessage):
    key: bytes


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class LocalSyncServerMessage(LocalSyncMessage):
    """Message from the backend to send to the client."""


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class NewProjectMessage(LocalSyncServerMessage):
    project_name: str
    project_actor_id: str


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class GetProjectMessage(LocalSyncServerMessage):
    project: Optional[Project]


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class SyncBranchResponseMessage(LocalSyncServerMessage):
    pass


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class SyncBranchSuccessful(SyncBranchResponseMessage):
    pass


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class SyncBranchFailed(SyncBranchResponseMessage):
    remote_state_description: str


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class FailedToSaveMessage(LocalSyncServerMessage):
    version: int


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class AppliedSaveMessage(LocalSyncServerMessage):
    version: int
    is_containing_new_changes: bool


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class SecretsUpdatedMessage(LocalSyncServerMessage):
    pass


class SecretsUpdateFailureReason(StrEnum):
    NO_SECRET_KEY = "NO_SECRET_KEY"
    INVALID_SECRET_KEY = "INVALID_SECRET_KEY"
    UNKNOWN = "UNKNOWN"


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class SecretsFailedToUpdateMessage(LocalSyncServerMessage):
    reason: SecretsUpdateFailureReason


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class NewRemoteRepoStateMessage(LocalSyncServerMessage):
    branch_name: str
    version: int
    commit_hash: str
    repo_operation: Union[CleanRepoOperation, ConflictedRepoOperation] = attr.ib(converter=_convert_to_repo_operation)


@attr.s(auto_attribs=True, frozen=True, kw_only=True)
class SyncedMessage(NewRemoteRepoStateMessage):
    pass



import asyncio
from contextlib import asynccontextmanager
from typing import AsyncGenerator

import attr
import pytest

from imbue_core.async_utils import AsyncCachedProperty
from imbue_core.async_utils import sync_contextmanager


@pytest.mark.asyncio
async def test_async_cached_property() -> None:
    class Foo:
        def __init__(self) -> None:
            self.called_times = 0

        @AsyncCachedProperty
        async def bar(self) -> int:
            self.called_times += 1
            return 42

    foo = Foo()
    assert await foo.bar == 42
    assert await foo.bar == 42
    assert foo.called_times == 1


@pytest.mark.asyncio
async def test_clear_cached_property() -> None:
    class Foo:
        def __init__(self) -> None:
            self.called_times = 0

        @AsyncCachedProperty
        async def bar(self) -> int:
            self.called_times += 1
            return 42

    foo = Foo()
    assert await foo.bar == 42
    assert await foo.bar == 42
    assert foo.called_times == 1
    del foo.bar
    assert await foo.bar == 42
    assert foo.called_times == 2


async def returns1() -> int:
    return 1


@pytest.mark.asyncio
async def test_async_cached_property_set() -> None:
    class Foo:
        def __init__(self) -> None:
            self.called_times = 0

        @AsyncCachedProperty
        async def bar(self) -> int:
            self.called_times += 1
            return 42

    foo = Foo()
    foo.bar = 1

    assert await foo.bar == 1
    assert await foo.bar == 1
    assert foo.called_times == 0

    del foo.bar
    assert await foo.bar == 42
    assert foo.called_times == 1


@pytest.mark.asyncio
async def test_async_cached_property_on_frozen_attrs_class() -> None:
    @attr.s(auto_attribs=True, frozen=True)
    class Foo:
        x: int

        @AsyncCachedProperty
        async def bar(self) -> int:
            return self.x + 1

    foo = Foo(41)
    assert await foo.bar == 42
    assert await foo.bar == 42


@pytest.mark.asyncio
async def test_async_cached_property_concurrent() -> None:
    """Tests that the property is only computed once when accessed concurrently
    during the first computation."""

    class Foo:
        def __init__(self, in_bar_event: asyncio.Event, finish_bar_event: asyncio.Event) -> None:
            self._in_bar_event = in_bar_event
            self._finish_bar_event = finish_bar_event
            self.called_times = 0

        @AsyncCachedProperty
        async def bar(self) -> int:
            self.called_times += 1
            self._in_bar_event.set()
            await self._finish_bar_event.wait()
            return 42

    in_bar_event = asyncio.Event()
    finish_bar_event = asyncio.Event()
    foo = Foo(in_bar_event, finish_bar_event)

    async def bar() -> int:
        return await foo.bar

    task1 = asyncio.create_task(bar())
    task2 = asyncio.create_task(bar())

    await in_bar_event.wait()
    assert foo.called_times == 1
    finish_bar_event.set()
    r1, r2 = await asyncio.gather(task1, task2)
    assert r1 == 42
    assert r2 == 42
    assert foo.called_times == 1


def test_sync_contextmanager() -> None:
    state = []

    @asynccontextmanager
    async def async_contextmanager() -> AsyncGenerator[int, None]:
        state.append(1)
        await asyncio.sleep(0)
        try:
            yield 5
        except AssertionError as e:
            state.append(2)
        else:
            state.append(3)

    with sync_contextmanager(async_contextmanager()) as value:
        pass

    assert value == 5
    assert state == [1, 3]

    del value
    state = []

    with sync_contextmanager(async_contextmanager()) as value:
        raise AssertionError

    assert value == 5
    assert state == [1, 2]

    del value
    state = []
    with pytest.raises(ValueError):
        with sync_contextmanager(async_contextmanager()) as value:
            raise ValueError

    assert value == 5
    assert state == [1]



from abc import ABC
from abc import abstractmethod
from copy import deepcopy
from functools import cached_property
from typing import Any
from typing import Dict
from typing import FrozenSet
from typing import Iterable
from typing import List
from typing import Mapping
from typing import NoReturn
from typing import Protocol
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import TypeAlias
from typing import TypeVar
from typing import Union
from typing import cast


class _SupportsLessThan(Protocol):
    def __lt__(self, __other: Any) -> bool:
        ...


T = TypeVar("T")
TV = TypeVar("TV")
TK = TypeVar("TK", bound=_SupportsLessThan)


class FrozenMapping(Mapping[T, TV], ABC):
    @abstractmethod
    def __hash__(self) -> int:
        ...


class FrozenDict(Dict[T, TV], FrozenMapping[T, TV]):
    def _key(self) -> FrozenSet[Tuple[T, TV]]:
        return frozenset(self.items())

    @cached_property
    def _hash(self) -> int:
        # bawr said it should be fine
        return hash(self._key())

    def __hash__(self) -> int:  # type: ignore
        return self._hash

    def _mutation_error(self, method: str) -> RuntimeError:
        return RuntimeError(f"Cannot call mutation method {method} on _FrozenDict {self}")

    def __setitem__(self, __name: T, __value: TV) -> NoReturn:
        raise self._mutation_error("__setitem__")

    def __delitem__(self, __name: T) -> NoReturn:
        raise self._mutation_error("__delitem__")

    def update(self, __m: Mapping[T, TV]) -> NoReturn:  # type: ignore
        raise self._mutation_error("update")

    def setdefault(self, __name: T, __value: TV) -> NoReturn:
        raise self._mutation_error("setdefault")

    def pop(self, __name: T, __default: TV) -> NoReturn:  # type: ignore
        raise self._mutation_error("pop")

    def popitem(self) -> NoReturn:
        raise self._mutation_error("popitem")

    def clear(self) -> NoReturn:
        raise self._mutation_error("clear")

    def __repr__(self) -> str:
        return f"_FrozenDict({super().__repr__()})"

    def __copy__(self) -> "FrozenDict":
        return type(self)(self)

    def __deepcopy__(self, memo: Dict[int, Any]) -> "FrozenDict":
        memo[id(self)] = self
        copied_items = ((deepcopy(key, memo), deepcopy(value, memo)) for key, value in self.items())
        return type(self)(copied_items)

    def __reduce__(self) -> Tuple[Any, ...]:
        return (FrozenDict, (dict(self),))


def empty_mapping() -> FrozenDict[Any, Any]:
    return FrozenDict()


def deep_freeze_mapping(mapping: Mapping[T, TV]) -> FrozenDict[T, Any]:
    return FrozenDict({key: cast(TV, _deep_freeze_any(value)) for key, value in mapping.items()})


def _freeze_iterable_values(iterable: Iterable[T]) -> Iterable[Any]:
    return (cast(T, _deep_freeze_any(value)) for value in iterable)


def deep_freeze_set(input_set: Union[Set[T], FrozenSet[T]]) -> FrozenSet[Any]:
    return frozenset(_freeze_iterable_values(input_set))


def _deep_freeze_any(input_object: object) -> object:
    if isinstance(input_object, Mapping):
        return deep_freeze_mapping(input_object)

    if isinstance(input_object, (set, frozenset)):
        return deep_freeze_set(input_object)

    if (
        isinstance(input_object, Iterable)
        and not isinstance(input_object, str)
        and not isinstance(input_object, bytes)
    ):
        return tuple(_freeze_iterable_values(input_object))

    return input_object


def deep_freeze_sequence(sequence: Sequence[T]) -> Tuple[Any, ...]:
    return tuple(_freeze_iterable_values(sequence))


# Recursive type alias that captures the possible types of JSON objects (e.g. from json.loads).
JSON: TypeAlias = Union[str, int, bool, float, None, Dict[str, "JSON"], List["JSON"]]


# Immutable version of JSON.
FrozenJSON: TypeAlias = Union[str, int, bool, float, None, FrozenDict[str, "FrozenJSON"], Tuple["FrozenJSON", ...]]


def deep_freeze_json(json: JSON) -> FrozenJSON:
    if isinstance(json, dict):
        return FrozenDict({k: deep_freeze_json(v) for k, v in json.items()})
    elif isinstance(json, list):
        return tuple(deep_freeze_json(v) for v in json)
    else:
        return json
